-- date: 16.03.2015
-- author: rafftnix
-- version: v 2.0 (LS 17)

-- sourcecode for the Krone Ultima CF 155XC

Ultima = {}
Ultima.modDir = g_currentModDirectory;
Ultima.STATE_usualLoading = 1;
Ultima.STATE_netBinding = 2;
Ultima.STATE_dropBale = 3;
Ultima.STATE_passLoad = 4;

Ultima.WRAPPERSTATE_empty = 1;
Ultima.WRAPPERSTATE_tiltTable = 2;
Ultima.WRAPPERSTATE_fetchBale = 3;
Ultima.WRAPPERSTATE_wrapping = 4;
Ultima.WRAPPERSTATE_dropCheck = 5;
Ultima.WRAPPERSTATE_dropBale = 6;

Ultima.DROPMODE_AUTO = 1;
Ultima.DROPMODE_MANUAL = 2;
Ultima.DROPMODE_COLLECT = 3;

function Ultima.prerequisitesPresent(specializations)
    return SpecializationUtil.hasSpecialization(Attachable, specializations) and SpecializationUtil.hasSpecialization(Fillable, specializations) and SpecializationUtil.hasSpecialization(Pickup, specializations);
end;

function Ultima:load(savegame)
    self.setBalerState = SpecializationUtil.callSpecializationsFunction("setBalerState");
    self.setWrapperState = SpecializationUtil.callSpecializationsFunction("setWrapperState");
    self.setBaleSize = SpecializationUtil.callSpecializationsFunction("setBaleSize");
    self.setBaleDropMode = SpecializationUtil.callSpecializationsFunction("setBaleDropMode");
    self.setWrapperActive = SpecializationUtil.callSpecializationsFunction("setWrapperActive");
    self.setBaleCounter = SpecializationUtil.callSpecializationsFunction("setBaleCounter");
    self.setNetRoleOnStoragePlace = SpecializationUtil.callSpecializationsFunction("setNetRoleOnStoragePlace");
    self.setNetRoleFillState = SpecializationUtil.callSpecializationsFunction("setNetRoleFillState");
    self.foilStorageSetRole = SpecializationUtil.callSpecializationsFunction("foilStorageSetRole");
    self.setWrapFoilLength = SpecializationUtil.callSpecializationsFunction("setWrapFoilLength");
    self.playerTakeRole = SpecializationUtil.callSpecializationsFunction("playerTakeRole");
    self.playerRemoveRole = SpecializationUtil.callSpecializationsFunction("playerRemoveRole");    
    self.wrapperDeleteOldBale = SpecializationUtil.callSpecializationsFunction("wrapperDeleteOldBale");
    self.linkBaleToWrapper = SpecializationUtil.callSpecializationsFunction("linkBaleToWrapper");
    self.balerCreateBale = Ultima.balerCreateBale;
    self.wrapperCreateBale = SpecializationUtil.callSpecializationsFunction("wrapperCreateBale");
    self.wrapperDropBale = SpecializationUtil.callSpecializationsFunction("wrapperDropBale");
    self.allowPickingUp = Utils.overwrittenFunction(self.allowPickingUp, Ultima.allowPickingUp);
    self.setPickupState = Utils.overwrittenFunction(self.setPickupState, Ultima.setPickupState);
    self.getDirtMultiplier = Utils.overwrittenFunction(self.getDirtMultiplier, Ultima.getDirtMultiplier);
    self.loadSpeedRotatingPartFromXML = Utils.overwrittenFunction(self.loadSpeedRotatingPartFromXML, ForageWagon.loadSpeedRotatingPartFromXML);
    self.getIsSpeedRotatingPartActive = Utils.overwrittenFunction(self.getIsSpeedRotatingPartActive, ForageWagon.getIsSpeedRotatingPartActive);
    self.doCheckSpeedLimit = Utils.overwrittenFunction(self.doCheckSpeedLimit, Ultima.doCheckSpeedLimit);
    self.getRemainingAnimationTime = Ultima.getRemainingAnimationTime;   
    self.dropmatRaycast = Ultima.dropmatRaycast;    
    self.passLoadToMainChamber = Ultima.passLoadToMainChamber;    
    self.setUnitFillLevel = Utils.overwrittenFunction(self.setUnitFillLevel, Ultima.setUnitFillLevel);
    self.allowFillType = Ultima.allowFillType;
    self.getMPCFillType = Ultima.getMPCFillType;
    self.getPPCFillType = Ultima.getPPCFillType;
                
    --self.usualLoadingStart = SpecializationUtil.callSpecializationsFunction("usualLoadingStart");
    self.usualLoadingUpdate = SpecializationUtil.callSpecializationsFunction("usualLoadingUpdate");
    self.netBindingStart = SpecializationUtil.callSpecializationsFunction("netBindingStart");
    self.netBindingUpdate = SpecializationUtil.callSpecializationsFunction("netBindingUpdate");
    self.dropBaleStart = SpecializationUtil.callSpecializationsFunction("dropBaleStart");
    self.dropBaleUpdate = SpecializationUtil.callSpecializationsFunction("dropBaleUpdate");
    self.passLoadStart = SpecializationUtil.callSpecializationsFunction("passLoadStart");
    self.passLoadUpdate = SpecializationUtil.callSpecializationsFunction("passLoadUpdate");
    
    --self.wrapperEmptyStart = SpecializationUtil.callSpecializationsFunction("wrapperEmptyStart");
    --self.wrapperEmptyUpdate = SpecializationUtil.callSpecializationsFunction("wrapperEmptyUpdate");
    self.wrapperTiltTableStart = SpecializationUtil.callSpecializationsFunction("wrapperTiltTableStart");
    self.wrapperTiltTableUpdate = SpecializationUtil.callSpecializationsFunction("wrapperTiltTableUpdate");
    self.wrapperFetchBaleStart = SpecializationUtil.callSpecializationsFunction("wrapperFetchBaleStart");
    self.wrapperFetchBaleUpdate = SpecializationUtil.callSpecializationsFunction("wrapperFetchBaleUpdate");
    self.wrapperWrappingStart = SpecializationUtil.callSpecializationsFunction("wrapperWrappingStart");
    self.wrapperWrappingUpdate = SpecializationUtil.callSpecializationsFunction("wrapperWrappingUpdate");
    self.wrapperDropCheckStart = SpecializationUtil.callSpecializationsFunction("wrapperDropCheckStart");
    self.wrapperDropCheckUpdate = SpecializationUtil.callSpecializationsFunction("wrapperDropCheckUpdate");
    self.wrapperDropBaleStart = SpecializationUtil.callSpecializationsFunction("wrapperDropBaleStart");
    self.wrapperDropBaleUpdate = SpecializationUtil.callSpecializationsFunction("wrapperDropBaleUpdate");
        
    self.currentBalerState = 1;
    self.currentWrapperState = 1;
    self.currentBaleDropMode = 1;
    self.wrapperIsActive = true;
    self.dropNextBaleDirectly = false;
    self.wrapperBaleSizeIndex = 1;
    self.lastTrackingPos = {0, 0, 0} -- average input tracking
    self.averageInputPerM = 1; -- prevent dividing with 0
    self.inputTracker = {}
    self.trackDistance = 5;
    self.remainingAnimTimeBuffer = 250; 
    self.speedLimitBufferNum = 0;
    self.speedLimitBufferAmount = 0;
    self.maxSpeedLimit = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.speedLimit#value"), 15);
    self.UVScrollParts = Utils.loadScrollers(self.components, self.xmlFile, "vehicle.Ultima.uvScrollShader.UVScrollPart", {}, false);
    
    self.mainChamberUnitIndex = getXMLFloat(self.xmlFile, "vehicle.Ultima.fillUnits#mainChamberFillUnitIndex");
    self.preChamberUnitIndex = getXMLFloat(self.xmlFile, "vehicle.Ultima.fillUnits#prePressChamberFillUnitIndex");
    
    self.ppcPassThrough = Utils.getNoNil(getXMLFloat(self.xmlFile, "vehicle.Ultima.prePressChamber#literPerSecond"), 100);
    
    self.baleCountDay = 0;
    self.baleCountLifetime = 0;
    
    self.netBindingTimer = 0;
    self.netBindingTimerTime = 2000; -- dynamic?
    
    self.ultimaTurnOnRotNodes = Utils.loadRotationNodes(self.xmlFile, {}, "vehicle.turnedOnRotationNodes.turnedOnRotationNode", "baler", self.components);

    -- baler states
    self.balerStates = {}
    function registerBalerState(stateId, startFunction, updateFunction)
        local state = {}
        state.stateId = stateId;
        state.startFunction = startFunction;
        state.updateFunction = updateFunction;
        self.balerStates[stateId] = state;
    end;
    registerBalerState(Ultima.STATE_usualLoading, self.usualLoadingStart, self.usualLoadingUpdate);
    registerBalerState(Ultima.STATE_netBinding, self.netBindingStart, self.netBindingUpdate);
    registerBalerState(Ultima.STATE_dropBale, self.dropBaleStart, self.dropBaleUpdate);
    registerBalerState(Ultima.STATE_passLoad, self.passLoadStart, self.passLoadUpdate);

    -- wrapper states
    self.wrapperStates = {}
    function registerWrapperState(stateId, startFunction, updateFunction)
        local state = {}
        state.stateId = stateId;
        state.startFunction = startFunction;
        state.updateFunction = updateFunction;
        self.wrapperStates[stateId] = state;
    end;
    registerWrapperState(Ultima.WRAPPERSTATE_empty, self.wrapperEmptyStart, self.wrapperEmptyUpdate);
    registerWrapperState(Ultima.WRAPPERSTATE_tiltTable, self.wrapperTiltTableStart, self.wrapperTiltTableUpdate);
    registerWrapperState(Ultima.WRAPPERSTATE_fetchBale, self.wrapperFetchBaleStart, self.wrapperFetchBaleUpdate);
    registerWrapperState(Ultima.WRAPPERSTATE_wrapping, self.wrapperWrappingStart, self.wrapperWrappingUpdate);
    registerWrapperState(Ultima.WRAPPERSTATE_dropCheck, self.wrapperDropCheckStart, self.wrapperDropCheckUpdate);
    registerWrapperState(Ultima.WRAPPERSTATE_dropBale, self.wrapperDropBaleStart, self.wrapperDropBaleUpdate);
    
    -- register i3d animations
    -- times are absolute
    function registerI3dAnimation(xmlFile, xmlPath, loop)
        local i3dAnim = {}
        i3dAnim.nodeId = Utils.indexToObject(self.components, getXMLString(self.xmlFile, xmlPath.."#nodeIndex"));
        i3dAnim.charSet = getAnimCharacterSet(i3dAnim.nodeId);
        local clip = getAnimClipIndex(i3dAnim.charSet, getXMLString(self.xmlFile, xmlPath.."#animName"));
        i3dAnim.speedScale = Utils.getNoNil(getXMLFloat(self.xmlFile, xmlPath.."#speedScale"), 1);
        i3dAnim.trackIndex = Utils.getNoNil(getXMLFloat(self.xmlFile, xmlPath.."#trackIndex"), 0);
        i3dAnim.duration = getXMLFloat(self.xmlFile, xmlPath.."#duration");
    
        assignAnimTrackClip(i3dAnim.charSet, i3dAnim.trackIndex, clip);
        setAnimTrackLoopState(i3dAnim.charSet, i3dAnim.trackIndex, loop);
        setAnimTrackSpeedScale(i3dAnim.charSet, i3dAnim.trackIndex, i3dAnim.speedScale);
        
        function i3dAnim:play(absStartTime)
            startTime = Utils.getNoNil(startTime, 0);
            setAnimTrackTime(i3dAnim.charSet, i3dAnim.trackIndex, startTime);
            enableAnimTrack(i3dAnim.charSet, i3dAnim.trackIndex);
        end;
        
        function i3dAnim:stop()
            disableAnimTrack(i3dAnim.charSet, i3dAnim.trackIndex);
        end;
        
        function i3dAnim:getTime()
            return getAnimTrackTime(i3dAnim.charSet, i3dAnim.trackIndex);
        end;
        
        function i3dAnim:setTime(trackTime, immediateUpdate)
            setAnimTrackTime(i3dAnim.charSet, i3dAnim.trackIndex, trackTime, immediateUpdate);
        end;
        
        function i3dAnim:setRelativeTime(relTime)
            i3dAnim:setTime(relTime*i3dAnim.duration);
        end;
        
        function i3dAnim:setSpeedScale(scale)
            i3dAnim.speedScale = scale;
            setAnimTrackSpeedScale(i3dAnim.charSet, i3dAnim.trackIndex, i3dAnim.speedScale);
        end;
        
        function i3dAnim:getIsPlaying()
            return isAnimTrackEnabled(i3dAnim.charSet, i3dAnim.trackIndex);
        end;
        
        return i3dAnim;
    end;
    
    -- chamber fill animation
    self.chamberFillAnim = registerI3dAnimation(self.xmlFile, "vehicle.Ultima.baleSizes", false);

    self.balerBaleConductNode = Utils.indexToObject(self.components, getXMLString(self.xmlFile, "vehicle.Ultima.balerBaleConductNode#index"));
    self.balerBaleConductNodeRot = Utils.getVectorNFromString(getXMLString(self.xmlFile, "vehicle.Ultima.balerBaleConductNode#rotSpeed"), 3);
    self.wrapperBaleConductNode = Utils.indexToObject(self.components, getXMLString(self.xmlFile, "vehicle.Ultima.balerBaleConductNode#wrapperBaleConductNodeIndex"));
    
    self.playerRolePos = Utils.getVectorNFromString(getXMLString(self.xmlFile, "vehicle.Ultima.foilRoleStorages#playerRolePos"), 3);
    self.playerRoleRot = Utils.getRadiansFromString(getXMLString(self.xmlFile, "vehicle.Ultima.foilRoleStorages#playerRoleRot"), 3);
    self.foilRoleCloneNode = Utils.indexToObject(self.components, getXMLString(self.xmlFile, "vehicle.Ultima.foilRoleStorages#cloneNodeIndex"));
    
    self.wrapperRolesAnim = registerI3dAnimation(self.xmlFile, "vehicle.Ultima.wrapperRollsAnim", false);
    
    -- bale sizes and fillTypes
    self.baleSizeIndex = 1;
    self.baleSizes = {}
    local num = 0;
    while true do
        local key = "vehicle.Ultima.baleSizes.baleSize("..tostring(num)..")";
        if not hasXMLProperty(self.xmlFile, key) then
            break;
        end;
        local baleSize = {}
        baleSize.size = getXMLFloat(self.xmlFile, key.."#size")/100;
        baleSize.fillLevel = getXMLFloat(self.xmlFile, key.."#fillLevel");
        baleSize.chamberAnimTime = getXMLFloat(self.xmlFile, key.."#chamberAnimTime");
        
        baleSize.dropBaleAnimation = registerI3dAnimation(self.xmlFile, key..".dropBaleAnimation", false);
        baleSize.receiveBaleAnimTime = getXMLFloat(self.xmlFile, key..".dropBaleAnimation#receiveBaleAnimTime");
        baleSize.wrapperReceiveAnim = getXMLString(self.xmlFile, key..".wrapperReceiveBaleAnimation#xmlAnimName");
        baleSize.wrapAnimation = registerI3dAnimation(self.xmlFile, key..".wrapAnimation", false);
        local foilCutIndex1 = Utils.indexToObject(self.components, getXMLString(self.xmlFile, key..".wrapAnimation#wrapFoilCutIndex1"));
        local foilCutIndex2 = Utils.indexToObject(self.components, getXMLString(self.xmlFile, key..".wrapAnimation#wrapFoilCutIndex2"));
        baleSize.wrapFoilCuts = {foilCutIndex1, foilCutIndex2}
        baleSize.createPhysicalBaleTime = getXMLFloat(self.xmlFile, key..".wrapAnimation#createPhysicalBaleTime");
        baleSize.wrapperDropAnimName = getXMLString(self.xmlFile, key..".wrapperDropAnimation#xmlAnimName");
        baleSize.wrapAnimation:play(0); 
        baleSize.wrapAnimation:setTime(0, true);
        baleSize.wrapAnimation:stop();
        
        baleSize.fillTypes = {}
        local lastDesc;
        local num2 = 0;
        while true do 
            local key2 = "vehicle.Ultima.baleSizes.baleSize("..tostring(num)..").fillTypes.fillType("..tostring(num2)..")";
            if not hasXMLProperty(self.xmlFile, key2) then
                break;
            end;
            local desc = {}
            desc.fillType = FillUtil.getFillTypesByNames(getXMLString(self.xmlFile, key2.."#Ftype"), "Warning: '"..self.configFileName.. "' has invalid fillType '%s'.")[1];
            desc.fileName = getXMLString(self.xmlFile, key2.."#fileName");
            if getXMLString(self.xmlFile, key2.."#wrappedFileName") ~= nil then
                desc.isWrapAble = true;
                desc.wrappedFileName = getXMLString(self.xmlFile, key2.."#wrappedFileName");
                desc.fileNameForWrapping = getXMLString(self.xmlFile, key2.."#fileNameForWrapping");
            else
                desc.isWrapAble = false;
            end;
            baleSize.fillTypes[desc.fillType] = desc; 
            lastDesc = desc;
            num2 = num2 + 1;
        end;        
        
        -- set the last given bale for all other fillTypes that may have been added
        for k, v in pairs(self:getUnitFillTypes(self.mainChamberUnitIndex)) do
            if baleSize.fillTypes[k] == nil and k ~= FillUtil.FILLTYPE_UNKNOWN and FruitUtil.fillTypeIsWindrow[k] and v then
                baleSize.fillTypes[k] = lastDesc;
            end;    
        end;
        
        baleSize.numFillTypes = num2;
        table.insert(self.baleSizes, baleSize);
        num = num + 1;
    end;
    
    self.balerFillTypes = {}
    for k, v in pairs(self:getUnitFillTypes(self.mainChamberUnitIndex)) do
        if k ~= FillUtil.FILLTYPE_UNKNOWN and FruitUtil.fillTypeIsWindrow[k] and v then
            table.insert(self.balerFillTypes, k);
        end;    
    end;
    
    -- side doors
    self.sideDoors = {}
    self.numSideDoors = 0;
    while true do
        local key = "vehicle.Ultima.sideDoors.sideDoor("..tostring(self.numSideDoors)..")";
        if not hasXMLProperty(self.xmlFile, key) then
            break;
        end;
        
        self.numSideDoors = self.numSideDoors + 1;
        local sideDoor = {}
        sideDoor.animName = getXMLString(self.xmlFile, key.."#animName");
        sideDoor.index = self.numSideDoors;
        sideDoor.vehicle = self;
        sideDoor.triggerId = Utils.indexToObject(self.components, getXMLString(self.xmlFile, key.."#triggerIndex"));
        sideDoor.playerInTrigger = false;
        addTrigger(sideDoor.triggerId, "triggerCallback", sideDoor);
        
        function sideDoor:open()
            sideDoor.vehicle:playAnimation(sideDoor.animName, 1, 0, false);
        end;
        function sideDoor:close()
            sideDoor.vehicle:playAnimation(sideDoor.animName, -1, 1, false);
        end;
        
        function sideDoor:getIsClosed()
            return sideDoor.vehicle:getAnimationTime(sideDoor.animName) == 0;
        end;
        function sideDoor:getIsOpened()
            return sideDoor.vehicle:getAnimationTime(sideDoor.animName) == 1;
        end;
        
        function sideDoor:triggerCallback(triggerId, otherId, onEnter, onLeave, onStay, otherShapeId)
            if g_currentMission.player ~= nil and otherId == g_currentMission.player.rootNode then
                if onEnter then
                    sideDoor.playerInTrigger = true;
                elseif onLeave then
                    sideDoor.playerInTrigger = false;
                end;
            end;
        end;
        
        self.sideDoors[sideDoor.index] = sideDoor;
        self.sideDoors[sideDoor.animName] = sideDoor;
    end;    
    
    -- dropmat
    if hasXMLProperty(self.xmlFile, "vehicle.Ultima.dropmat") then
        local key = "vehicle.Ultima.dropmat";
        self.dropmat = {}
        self.dropmat.triggerId = Utils.indexToObject(self.components, getXMLString(self.xmlFile, key.."#playerTriggerIndex"));
        self.dropmat.animName = getXMLString(self.xmlFile, key.."#foldAnimName");
        self.dropmat.raycastNode = Utils.indexToObject(self.components, getXMLString(self.xmlFile, key.."#raycastNodeIndex"));
        self.dropmat.jointNode = Utils.indexToObject(self.components, getXMLString(self.xmlFile, key.."#jointIndex"));
        self.dropmat.maxJointY = getXMLFloat(self.xmlFile, key.."#topJoint");
        self.dropmat.minJointY = getXMLFloat(self.xmlFile, key.."#bottomJoint");
        self.dropmat.playerInTrigger = false;
        self.dropmat.isFolding = false;
        function self.dropmat:triggerCallback(triggerId, otherId, onEnter, onLeave, onStay, otherShapeId)
            if g_currentMission.player ~= nil and otherId == g_currentMission.player.rootNode then
                if onEnter then
                    self.playerInTrigger = true;
                end;
                if onLeave then
                    self.playerInTrigger = false;
                end;
            end;        
        end;
        addTrigger(self.dropmat.triggerId, "triggerCallback", self.dropmat);
    end;
    
    -- net roles
    function registerNetRoleStorage(name, index)
        local storage = {}
        local key = "vehicle.Ultima.netRoleStorages."..name.."#nodeIndex";
        storage.nodeId = Utils.indexToObject(self.components, getXMLString(self.xmlFile, key));
        storage.index = index;
        return storage;
    end;
    
    self.netRoleStorageLeft = registerNetRoleStorage("leftRole", 1);
    self.netRoleStorageMiddle = registerNetRoleStorage("middleRole", 2);
    self.netRoleStorageRight = registerNetRoleStorage("rightRole", 3);
    self.netRoleStorageMiddle.refillAnimName = getXMLString(self.xmlFile, "vehicle.Ultima.netRoleStorages.middleRole#refillAnimationName");
    self.netRoleStorages = {self.netRoleStorageLeft, self.netRoleStorageMiddle, self.netRoleStorageRight}
    self.netRoleDefaultLenght = getXMLFloat(self.xmlFile, "vehicle.Ultima.netRoleStorages#defaultLenght")
    -- default: all net roles filled

    self.netRoleTop = {}
    self.netRoleTop.netRole = Utils.indexToObject(self.components, getXMLString(self.xmlFile, "vehicle.Ultima.netRoleStorages.topRole#netRoleIndex"));
    self.netRoleTop.role = Utils.indexToObject(self.components, getXMLString(self.xmlFile, "vehicle.Ultima.netRoleStorages.topRole#roleIndex"));
    self.netRoleTop.netRoleRotSpeed = Utils.getVectorNFromString(getXMLString(self.xmlFile, "vehicle.Ultima.netRoleStorages.topRole#netRoleRotSpeed"), 3);
    self.netRoleTop.fillAnim = getXMLString(self.xmlFile, "vehicle.Ultima.netRoleStorages.topRole#fillAnim");
    self.netRoleTop.length = self.netRoleDefaultLenght;
    self.netRoleTop.maxLength = self.netRoleDefaultLenght;
    self.numNetWraps = 4;
    self:setNetRoleFillState(self.netRoleTop.length);
    
    local num = 0;
    self.pickupEffectActiveTime = 0;
    self.pickupEffectActive = false;
    self.pickupEffect = EffectManager:loadEffect(self.xmlFile, "vehicle.Ultima.pickupEffect", self.components, self);
    
    local foilWrapMesh1 = Utils.indexToObject(self.components, getXMLString(self.xmlFile, "vehicle.Ultima.foilWrapMesh#index1"));
    local foilWrapMesh2 = Utils.indexToObject(self.components, getXMLString(self.xmlFile, "vehicle.Ultima.foilWrapMesh#index2"));
    self.foilWrapMesh = {foilWrapMesh1, foilWrapMesh2}
    
    local mesh1 = Utils.indexToObject(self.components, getXMLString(self.xmlFile, "vehicle.Ultima.foilWrapCutMesh#index1"));
    local mesh2 = Utils.indexToObject(self.components, getXMLString(self.xmlFile, "vehicle.Ultima.foilWrapCutMesh#index2"));
    self.foilWrapCutMesh = {mesh1, mesh2}
    
    -- foil roles
    self.foilRoleStorages = {}
    for a=1, 3 do
        local storage = {}
        storage.animName = getXMLString(self.xmlFile, "vehicle.Ultima.foilRoleStorages.storage("..tostring(a-1)..")#animName");
        storage.holders = {}
        storage.index = a;
        local num = 0;
        while true do
            local key = "vehicle.Ultima.foilRoleStorages.storage("..tostring(a-1)..").roleHolder("..tostring(num)..")#index";
            if not hasXMLProperty(self.xmlFile, key) then
                break;
            end;
            local nodeId = Utils.indexToObject(self.components, getXMLString(self.xmlFile, key));
            if not self.manualFoilRoleRefill then
                setVisibility(nodeId, true);
            end;
            table.insert(storage.holders, nodeId);
            num = num + 1;
        end;
        self.foilRoleStorages[a] = storage;
        self.foilRoleStorages[storage.animName] = storage;
    end;
    
    self.maxFoilLength = getXMLFloat(self.xmlFile, "vehicle.Ultima.wrapperFoilHolders#foilLength");
    self.baseFoilUsePerBale = getXMLFloat(self.xmlFile, "vehicle.Ultima.wrapperFoilHolders#baseFoilUsePerBale");
    self.wrapperFoilHolders = {}
    for a=1, 2 do
        local key = "vehicle.Ultima.wrapperFoilHolders.foilHolder("..tostring(a-1)..")";
        local foilHolder = {}
        foilHolder.nodeId = Utils.indexToObject(self.components, getXMLString(self.xmlFile, key.."#index"));
        foilHolder.playerTrigger = Utils.indexToObject(self.components, getXMLString(self.xmlFile, key.."#playerTriggerIndex"));
        foilHolder.emptyAnimName = getXMLString(self.xmlFile, key.."#emptyAnimName");
        foilHolder.playerInTrigger = false;
        foilHolder.remainingFoilLength = self.maxFoilLength;

        function foilHolder:triggerCallback(triggerId, otherId, onEnter, onLeave, onStay, otherShapeId)
            if g_currentMission.player ~= nil and g_currentMission.player.rootNode == otherId then
                if onEnter then
                    foilHolder.playerInTrigger = true;
                end;
                if onLeave then
                    foilHolder.playerInTrigger = false;
                end;
            end;
        end;
        addTrigger(foilHolder.playerTrigger, "triggerCallback", foilHolder);
        table.insert(self.wrapperFoilHolders, foilHolder);
    end;
    
    -- sounds 
    local novoGripSoundNode = Utils.indexToObject(self.components, getXMLString(self.xmlFile, "vehicle.Ultima.sounds.wrapping#nodeIndex"));
    local wrappingSoundNode = Utils.indexToObject(self.components, getXMLString(self.xmlFile, "vehicle.Ultima.sounds.novoGrip#nodeIndex"));
    self.wrappingStartSound = SoundUtil.loadSample(self.xmlFile, {}, "vehicle.Ultima.sounds.wrappingStart", nil, self.baseDirectory);
    self.wrappingStopSound = SoundUtil.loadSample(self.xmlFile, {}, "vehicle.Ultima.sounds.wrappingStop", nil, self.baseDirectory);
    self.wrapping3DSound = SoundUtil.loadSample(self.xmlFile, {}, "vehicle.Ultima.sounds.wrapping", nil, self.baseDirectory, wrappingSoundNode);
    self.baleDoorOpenSound = SoundUtil.loadSample(self.xmlFile, {}, "vehicle.Ultima.sounds.baleDoorOpen", nil, self.baseDirectory);
    self.baleDoorCloseSound = SoundUtil.loadSample(self.xmlFile, {}, "vehicle.Ultima.sounds.baleDoorClose", nil, self.baseDirectory);
    self.novoGripStartSound = SoundUtil.loadSample(self.xmlFile, {}, "vehicle.Ultima.sounds.novoGripStart", nil, self.baseDirectory);
    self.novoGripStopSound = SoundUtil.loadSample(self.xmlFile, {}, "vehicle.Ultima.sounds.novoGripStop", nil, self.baseDirectory);
    self.novoGrip3DSound = SoundUtil.loadSample(self.xmlFile, {}, "vehicle.Ultima.sounds.novoGrip", nil, self.baseDirectory, novoGripSoundNode);
    self.sideDoorOpenSound = SoundUtil.loadSample(self.xmlFile, {}, "vehicle.Ultima.sounds.sideDoorOpen", nil, self.baseDirectory);
    self.sideDoorCloseSound = SoundUtil.loadSample(self.xmlFile, {}, "vehicle.Ultima.sounds.sideDoorClose", nil, self.baseDirectory);
    self.netBindingStartSound = SoundUtil.loadSample(self.xmlFile, {}, "vehicle.Ultima.sounds.netBindingStart", nil, self.baseDirectory);
    self.netBindingStopSound = SoundUtil.loadSample(self.xmlFile, {}, "vehicle.Ultima.sounds.netBindingStop", nil, self.baseDirectory);
    self.pickupUpSound = SoundUtil.loadSample(self.xmlFile, {}, "vehicle.Ultima.sounds.pickupUp", nil, self.baseDirectory);
    self.pickupDownSound = SoundUtil.loadSample(self.xmlFile, {}, "vehicle.Ultima.sounds.pickupDown", nil, self.baseDirectory);
    self.wrapperTiltTableSound = SoundUtil.loadSample(self.xmlFile, {}, "vehicle.Ultima.sounds.wrapperTiltTable", nil, self.baseDirectory);
    self.wrapperDropBaleSound = SoundUtil.loadSample(self.xmlFile, {}, "vehicle.Ultima.sounds.wrapperDropBale", nil, self.baseDirectory);
    self.balerAlarmSound = SoundUtil.loadSample(self.xmlFile, {}, "vehicle.Ultima.sounds.balerAlarm", nil, self.baseDirectory);
    
    self.hud = RafftnixGUI:new(nil, false, false);
    
    --local startX = g_currentMission.vehicleHudPosX;
    --local startY = g_currentMission.vehicleHudPosY + g_currentMission.vehicleBaseHudHeight + g_currentMission.hudBackgroundHeight;
    local uiScale = g_gameSettings:getValue("uiScale");
    
    --local startX = 1-0.19*uiScale;                    -- values @ uiScale 1.0
    --local startY = 0.21*uiScale;
    
    --local startX = 1-0.21*uiScale;                    -- values @ uiScale 0.5
    --local startY = 0.25*uiScale;
    
    local startX = 1-0.21*uiScale+(0.04*(uiScale-0.5));    
    local startY = 0.25*uiScale-(0.08*(uiScale-0.5));
    
    
    local HUDWidth = 1-startX;
    local iconWidth = HUDWidth/4; -- 4 icons -- icons are squares
    local iconHeight = iconWidth * g_screenAspectRatio;
    local textSize = 0.02*uiScale;
    
    self.ppcOpenImg = self.hud:createImage(self.hud.baseElement, startX, startY, iconWidth, iconHeight, "hud/F4_2.dds");
    self.ppcClosedImg = self.hud:createImage(self.hud.baseElement, startX, startY, iconWidth, iconHeight, "hud/F4_1.dds");

    self.netBindingActiveImg = self.hud:createImage(self.hud.baseElement, startX+iconWidth, startY, iconWidth, iconHeight, "hud/F1_2.dds");
    self.netBindingPassiveImg = self.hud:createImage(self.hud.baseElement, startX+iconWidth, startY, iconWidth, iconHeight, "hud/F1_1.dds");

    self.baleDoorOpeningImg = self.hud:createImage(self.hud.baseElement, startX+iconWidth*2, startY, iconWidth, iconHeight, "hud/F5_2.dds");
    self.baleDoorClosedImg = self.hud:createImage(self.hud.baseElement, startX+iconWidth*2, startY, iconWidth, iconHeight, "hud/F5_1.dds");
    
    self.wrapperActiveImg = self.hud:createImage(self.hud.baseElement, startX+iconWidth*3, startY, iconWidth, iconHeight, "hud/F6_2.dds");
    self.wrapperPassiveImg = self.hud:createImage(self.hud.baseElement, startX+iconWidth*3, startY, iconWidth, iconHeight, "hud/F6_1.dds");
    
    -- chamber fill states above icon row
    self.blackBackGroundImg = self.hud:createImage(self.hud.baseElement, startX, startY+iconHeight, HUDWidth, 0.065, "hud/balken.dds");
    self.mainPressChamberText = self.hud:createText(self.hud.baseElement, startX, startY+iconHeight+0.045, HUDWidth, 0, "", textSize, 1);
    self.prePressChamberText = self.hud:createText(self.hud.baseElement, startX, startY+iconHeight+0.015, HUDWidth, 0, "", textSize, 1);
    self.mainPressChamberText:setTextFocus("middle", "bottom");
    self.prePressChamberText:setTextFocus("middle", "bottom");
    
    local baleCounterY = startY + iconHeight + 0.065;
    local baleCounterSymbolWidth = 0.025*uiScale;
    local baleCounterHeight = baleCounterSymbolWidth * g_screenAspectRatio*uiScale;
    self.baleCounterAllBG = self.hud:createImage(self.hud.baseElement, 0.01+startX, baleCounterY, baleCounterSymbolWidth, baleCounterHeight, "hud/baleCounter_all.dds");
    self.baleCounterSingleBG = self.hud:createImage(self.hud.baseElement, startX + HUDWidth/2, baleCounterY, baleCounterSymbolWidth, baleCounterHeight, "hud/baleCounter_single.dds");
    self.baleCounterAllText = self.hud:createText(self.hud.baseElement, 0.01+startX+baleCounterSymbolWidth + baleCounterSymbolWidth/3, baleCounterY+baleCounterHeight/2, 0, 0, "0", baleCounterSymbolWidth, 1);
    self.baleCounterSingleText = self.hud:createText(self.hud.baseElement, startX+ HUDWidth/2 + baleCounterSymbolWidth + baleCounterSymbolWidth/3, baleCounterY+baleCounterHeight/2, 0, 0, "0", baleCounterSymbolWidth, 1);
    self.baleCounterAllText:setTextFocus("left", "center");
    self.baleCounterSingleText:setTextFocus("left", "center");
    
    local item = StoreItemsUtil.storeItemsByXMLFilename[self.configFileName:lower()];
    
    local configuration = item.configurations['refillNetRoles'][self.configurations['refillNetRoles']];
    self.manualNetRoleRefill = configuration.manualRefill;
    
    local configuration = item.configurations['refillFoilRoles'][self.configurations['refillFoilRoles']];
    self.manualFoilRoleRefill = configuration.manualRefill;
    
    -- black background for black jack
    self.blackBackGroundImg:setDimension(self.blackBackGroundImg.width, self.blackBackGroundImg.height+baleCounterHeight);

    if savegame ~= nil and not savegame.resetVehicles then
        local baleCountDay = Utils.getNoNil(getXMLInt(savegame.xmlFile, savegame.key.."#baleCountDay"), 0);
        local baleCountLifetime = Utils.getNoNil(getXMLInt(savegame.xmlFile, savegame.key.."#baleCountLifetime"), 0);
        self:setBaleCounter(baleCountDay, baleCountLifetime, true); 

        local wrapperIsActive = Utils.getNoNil(getXMLBool(savegame.xmlFile, savegame.key.."#wrapperIsActive"), true);
        self:setWrapperActive(wrapperIsActive, true);
        
        local baleSizeIndex = Utils.getNoNil(getXMLInt(savegame.xmlFile, savegame.key.."#baleSizeIndex"), 1);
        self:setBaleSize(baleSizeIndex, true);
        
        local baleDropMode = Utils.getNoNil(getXMLInt(savegame.xmlFile, savegame.key.."#baleDropMode"), 1);
        self:setBaleDropMode(baleDropMode, true);
        
        self.dropmat.isFolding = Utils.getNoNil(getXMLBool(savegame.xmlFile, savegame.key.."#dropmatIsFolding"), false);
        if self.dropmat.isFolding then
            self:playAnimation(self.dropmat.animName, 1, 0, true);
        else
            self:playAnimation(self.dropmat.animName, -1, 1, true);
        end;
        
        if self.manualNetRoleRefill then
            -- default: all net roles full, so we switch off the empty roles
            self:setNetRoleOnStoragePlace(1, Utils.getNoNil(getXMLBool(savegame.xmlFile, savegame.key..".netRoles#left"), false), true);
            self:setNetRoleOnStoragePlace(2, Utils.getNoNil(getXMLBool(savegame.xmlFile, savegame.key..".netRoles#middle"), false), true);
            self:setNetRoleOnStoragePlace(3, Utils.getNoNil(getXMLBool(savegame.xmlFile, savegame.key..".netRoles#right"), false), true);
        
            local netRoleLength = Utils.getNoNil(getXMLFloat(savegame.xmlFile, savegame.key..".netRoles#topRoleLength"), self.netRoleDefaultLenght);
            self:setNetRoleFillState(math.floor(netRoleLength));
        else    
            self:setNetRoleOnStoragePlace(1, true, true);
            self:setNetRoleOnStoragePlace(2, true, true);
            self:setNetRoleOnStoragePlace(3, true, true);
        end;
        
        if self.manualFoilRoleRefill then
            for a=1, 3 do
                local storage = self.foilRoleStorages[a];
                local holdersStr = getXMLString(savegame.xmlFile, savegame.key..".foilRoleStorages.storage"..tostring(a).."#holders");
                if holdersStr ~= nil and holdersStr ~= "" then
                    local holders = Utils.splitString(" ", holdersStr);
                    for b=1, table.getn(storage.holders) do
                        if holders[b] == "1" then
                            self:foilStorageSetRole(a, b, true, true);
                        else
                            self:foilStorageSetRole(a, b, false, true);
                        end;
                    end;
                end;
            end;
        end;
        
        local balerState = Utils.getNoNil(getXMLInt(savegame.xmlFile, savegame.key.."#balerState"), 1);
        --local wrapperState = Utils.getNoNil(getXMLInt(savegame.xmlFile, savegame.key.."#wrapperState"), 1);
        
        if self.manualFoilRoleRefill then
            self:setWrapFoilLength(1, Utils.getNoNil(getXMLFloat(savegame.xmlFile, savegame.key..".wrapperFoilHolders#role1"), self.maxFoilLength));
            self:setWrapFoilLength(2, Utils.getNoNil(getXMLFloat(savegame.xmlFile, savegame.key..".wrapperFoilHolders#role2"), self.maxFoilLength));
        end;
        
        --[[if hasXMLProperty(savegame.xmlFile, savegame.key..".tempBalerBale") then
            local sizeIndex = getXMLInt(savegame.xmlFile, savegame.key..".tempBalerBale#sizeIndex");
            local fillTypeStr = getXMLString(savegame.xmlFile, savegame.key..".tempBalerBale#fillType");
            local fillType = FillUtil.fillTypeNameToInt[fillTypeStr];
            if self.balerCurrentBale ~= nil then
                unlink(self.balerCurrentBale.node);
                delete(self.balerCurrentBale.node);
            end;
            self:setUnitCapacity(self.mainChamberUnitIndex, self.baleSizes[sizeIndex].fillLevel);
            self.balerCurrentBale = self:balerCreateBale(sizeIndex, fillType);
        end;
        if hasXMLProperty(savegame.xmlFile, savegame.key..".wrapperTempBale") then
            local sizeIndex = getXMLInt(savegame.xmlFile, savegame.key..".wrapperTempBale#sizeIndex");
            local fillTypeStr = getXMLString(savegame.xmlFile, savegame.key..".wrapperTempBale#fillType");
            local fillType = FillUtil.fillTypeNameToInt[fillTypeStr];
            self.wrapperCurrentBale = self:balerCreateBale(sizeIndex, fillType, true);
        elseif hasXMLProperty(savegame.xmlFile, savegame.key..".wrapperBaleObject") then
            local sizeIndex = getXMLInt(savegame.xmlFile, savegame.key..".wrapperBaleObject#sizeIndex");
            local firstFillTypeStr = getXMLString(savegame.xmlFile, savegame.key..".wrapperBaleObject#firstFillType");
            local balei3dFilename = getXMLString(savegame.xmlFile, savegame.key..".wrapperBaleObject#balei3dFilename");
            local firstFillType = FillUtil.fillTypeNameToInt[firstFillTypeStr];
            self.wrapperCurrentBale = self:balerCreateBale(sizeIndex, firstFillType, true);
            self:wrapperCreateBale(sizeIndex, firstFillType, balei3dFilename);
        end;]]
        
        self:setBalerState(balerState, true);
        --self:setWrapperState(wrapperState, true);
    end;
end;

function Ultima:delete()
    if self.hud ~= nil then
        self.hud:delete();
    end;

    for a=1, table.getn(self.sideDoors) do
        removeTrigger(self.sideDoors[a].triggerId);
    end;
    for a=1, table.getn(self.wrapperFoilHolders) do
        removeTrigger(self.wrapperFoilHolders[a].playerTrigger);
    end;
    removeTrigger(self.dropmat.triggerId);

    EffectManager:deleteEffects(self.pickupEffect);
    
    if self.wrapperCurrentBale ~= nil and self.wrapperCurrentBale.baleObject ~= nil then
        self.wrapperCurrentBale.baleObject:delete();
    end;
    SoundUtil.deleteSample(self.wrappingStartSound);
    SoundUtil.deleteSample(self.wrappingStopSound);
    SoundUtil.deleteSample(self.wrapping3DSound);
    SoundUtil.deleteSample(self.baleDoorOpenSound);
    SoundUtil.deleteSample(self.baleDoorCloseSound);
    SoundUtil.deleteSample(self.novoGripStartSound);
    SoundUtil.deleteSample(self.novoGripStopSound);
    SoundUtil.deleteSample(self.novoGrip3DSound);
    SoundUtil.deleteSample(self.sideDoorOpenSound);
    SoundUtil.deleteSample(self.sideDoorCloseSound);
    SoundUtil.deleteSample(self.netBindingStartSound);
    SoundUtil.deleteSample(self.netBindingStopSound);
    SoundUtil.deleteSample(self.pickupUpSound);
    SoundUtil.deleteSample(self.pickupDownSound);
    SoundUtil.deleteSample(self.wrapperDropBaleSound);
    SoundUtil.deleteSample(self.balerAlarmSound);
end;

function Ultima:getSaveAttributesAndNodes(nodeIdent)
    local attributes, nodes = "", "";
    
    attributes = attributes ..'baleCountDay="'..tostring(self.baleCountDay)..'" baleCountLifetime="'..tostring(self.baleCountLifetime)..'"';
    attributes = attributes ..' wrapperIsActive="'..tostring(self.wrapperIsActive)..'"';
    attributes = attributes ..' baleSizeIndex="'..tostring(self.baleSizeIndex)..'"';
    attributes = attributes ..' baleDropMode="'..tostring(self.currentBaleDropMode)..'"';
    attributes = attributes ..' balerState="'..tostring(self.currentBalerState)..'"';
    --attributes = attributes ..' wrapperState="'..tostring(self.currentWrapperState)..'"';
    attributes = attributes ..' dropmatIsFolding="'..tostring(self.dropmat.isFolding)..'"';
    
    if self.manualNetRoleRefill then
        nodes = nodes ..'        <netRoles left="'..tostring(getVisibility(self.netRoleStorageLeft.nodeId))..'" middle="'..tostring(getVisibility(self.netRoleStorageMiddle.nodeId))..'" right="'..tostring(getVisibility(self.netRoleStorageRight.nodeId))..'"';
        nodes = nodes ..' topRoleLength="'..tostring(self.netRoleTop.length)..'"/>\n';
    end;
    
    if self.manualFoilRoleRefill then
        nodes = nodes ..'        <foilRoleStorages>\n';
        for a=1, 3 do
            local storage = self.foilRoleStorages[a];
            nodes = nodes ..'            <storage'..tostring(a)..' holders="';
            for b=1, table.getn(storage.holders) do
                if getVisibility(storage.holders[b]) then
                    nodes = nodes .. "1 ";
                else
                    nodes = nodes .. "0 ";
                end;
            end;
            nodes = nodes ..'"/>\n';
        end;    
        nodes = nodes ..'        </foilRoleStorages>';

        nodes = nodes ..'\n        <wrapperFoilHolders role1="'..tostring(self.wrapperFoilHolders[1].remainingFoilLength)..'" role2="'..tostring(self.wrapperFoilHolders[2].remainingFoilLength)..'" />';
    end;
    
    --[[if self.balerCurrentBale ~= nil then
        local fillTypeName = FillUtil.fillTypeIndexToDesc[self.balerCurrentBale.fillType].name;
        nodes = nodes ..'\n         <tempBalerBale sizeIndex="'..tostring(self.balerCurrentBale.sizeIndex)..'" fillType="'..fillTypeName..'"/>';
    end;

    if self.wrapperCurrentBale ~= nil then
        if self.wrapperCurrentBale.baleObject == nil then
            local fillTypeName = FillUtil.fillTypeIndexToDesc[self.wrapperCurrentBale.fillType].name;
            nodes = nodes ..'\n         <wrapperTempBale sizeIndex="'..tostring(self.wrapperCurrentBale.sizeIndex)..'" fillType="'..fillTypeName..'"/>';
        else
            local fillTypeName = FillUtil.fillTypeIndexToDesc[self.wrapperCurrentBale.fillType].name;
            nodes = nodes ..'\n            <wrapperBaleObject sizeIndex="'..tostring(self.wrapperCurrentBale.sizeIndex)..'" firstFillType="'..tostring(fillTypeName)..'" balei3dFilename="'..tostring(self.wrapperCurrentBale.baleObject.baleI3dFilename)..'"/>';
        end;
    end;]]

    return attributes, nodes;
end;

function Ultima:writeStream(streamId, connection)
    if not connection:getIsServer() then 
        streamWriteInt8(streamId, self.currentBaleDropMode);
        streamWriteBool(streamId, self.wrapperIsActive);
        streamWriteInt8(streamId, self.baleSizeIndex);
        streamWriteInt32(streamId, self.baleCountDay);
        streamWriteInt32(streamId, self.baleCountLifetime);
        streamWriteBool(streamId, self.dropmat.isFolding);
        
        for a=1, table.getn(self.sideDoors) do
            local sideDoor = self.sideDoors[a];
            streamWriteBool(streamId, (sideDoor:getIsOpened() or (not sideDoor:getIsClosed() and self.animations[sideDoor.animName].currentSpeed > 0)));
        end;
        
        if self.manualNetRoleRefill then
            for a=1, table.getn(self.netRoleStorages) do
                streamWriteBool(streamId, getVisibility(self.netRoleStorages[a].nodeId));
            end;
        
            streamWriteInt32(streamId, math.floor(self.netRoleTop.length));
        end;
        
        if self.manualFoilRoleRefill then
            for a=1, 3 do
                local storage = self.foilRoleStorages[a];
                for b=1, table.getn(storage.holders) do
                    streamWriteBool(streamId, getVisibility(storage.holders[b]));
                end;
            end;
            
            streamWriteInt32(streamId, math.floor(self.wrapperFoilHolders[1].remainingFoilLength));
            streamWriteInt32(streamId, math.floor(self.wrapperFoilHolders[2].remainingFoilLength));
        end;
        
        if self.balerCurrentBale ~= nil then
            streamWriteBool(streamId, true);
            streamWriteInt8(streamId, self.balerCurrentBale.sizeIndex);
            streamWriteInt8(streamId, self.balerCurrentBale.fillType);    
        else
            streamWriteBool(streamId, false);
        end;
        
        if self.wrapperCurrentBale ~= nil then
            streamWriteBool(streamId, true);
            streamWriteInt8(streamId, self.wrapperCurrentBale.sizeIndex);
            streamWriteInt8(streamId, self.wrapperCurrentBale.fillType);    
            if self.wrapperCurrentBale.baleObject ~= nil then
                streamWriteBool(streamId, true);
            else
                streamWriteBool(streamId, false);
            end;
        else
            streamWriteBool(streamId, false);
        end;
        
        streamWriteInt8(streamId, self.currentBalerState);
        streamWriteInt8(streamId, self.currentWrapperState);
    end;
end;

function Ultima:readStream(streamId, connection)
    if connection:getIsServer() then
        self:setBaleDropMode(streamReadInt8(streamId), true);
        self:setWrapperActive(streamReadBool(streamId), true);
        self:setBaleSize(streamReadInt8(streamId), true);
        self:setBaleCounter(streamReadInt32(streamId), streamReadInt32(streamId), true);
        self.dropmat.isFolding = streamReadBool(streamId);

        if self.dropmat.isFolding then
            self:playAnimation(self.dropmat.animName, 1, 0, true);
        else
            self:playAnimation(self.dropmat.animName, -1, 1, true);
        end;
                
        for a=1, table.getn(self.sideDoors) do
            local sideDoor = self.sideDoors[a];
            if streamReadBool(streamId) then
                self:setAnimationTime(sideDoor.animName, 1, true);
            else
                self:setAnimationTime(sideDoor.animName, 0, true);
            end;
        end;
        
        if self.manualNetRoleRefill then
            for a=1, table.getn(self.netRoleStorages) do
                self:setNetRoleOnStoragePlace(a, streamReadBool(streamId), true);
            end;
        
            local netRoleLength = streamReadInt32(streamId);
            self:setNetRoleFillState(netRoleLength);
        end;
        
        if self.manualFoilRoleRefill then
            for a=1, 3 do
                local storage = self.foilRoleStorages[a];
                for b=1, table.getn(storage.holders) do
                    setVisibility(storage.holders[b], streamReadBool(streamId));
                end;
            end;
            
            for a=1, 2 do
                local length = streamReadInt32(streamId);
                self:setWrapFoilLength(a, length, true);
            end;
        end;
        
        if streamReadBool(streamId) then
            local sizeIndex = streamReadInt8(streamId);
            local fillType = streamReadInt8(streamId);
            self.balerCurrentBale = self:balerCreateBale(sizeIndex, fillType);
        end;
        
        if streamReadBool(streamId) then
            local sizeIndex = streamReadInt8(streamId);
            local fillType = streamReadInt8(streamId);
            self.wrapperCurrentBale = self:balerCreateBale(sizeIndex, fillType, true);
            if streamReadBool(streamId) then
                self:wrapperDeleteOldBale();
            end;
        end;
        
        self:setBalerState(streamReadInt8(streamId), true);
        self:setWrapperState(streamReadInt8(streamId), true);
    end;
end;

function Ultima:writeUpdateStream(streamId, connection, dirtyMask)
    if not connection:getIsServer() then
        streamWriteBool(streamId, self.pickupEffectActive);
    end;
end;

function Ultima:readUpdateStream(streamId, timestamp, connection)
    if connection:getIsServer() then
        self.pickupEffectActive = streamReadBool(streamId);
    end;
end;

function Ultima:mouseEvent(posX, posY, isDown, isUp, button)
end;

function Ultima:keyEvent(unicode, sym, modifier, isDown)
end;

function Ultima:update(dt)
    if self:getIsActiveForInput() then
        if self.baleCountDay > 0 then
            if InputBinding.hasEvent(InputBinding.IMPLEMENT_EXTRA4) then
                self:setBaleCounter(0, self.baleCountLifetime);
            end;
        end;
        if InputBinding.hasEvent(InputBinding.SETBALESIZE) then
            self:setBaleSize(self.baleSizeIndex + 1);
        end;
        if InputBinding.hasEvent(InputBinding.SETWRAPPERACTIVE) then
            self:setWrapperActive(not self.wrapperIsActive);
        end;
        if InputBinding.hasEvent(InputBinding.SETBALEDROPMODE) then
            self:setBaleDropMode(self.currentBaleDropMode+1);
        end;
        if InputBinding.hasEvent(InputBinding.HIDEGUI) then
            self.hud:setHideGUI(not self.hud.hideGUI);
        end;
        
        if self.currentBaleDropMode == Ultima.DROPMODE_MANUAL then
            if self.wrapperCurrentBale ~= nil and self.wrapperCurrentBale.baleObject ~= nil and self:getUnitFillLevel(self.mainChamberUnitIndex) > (self:getUnitCapacity(self.mainChamberUnitIndex) * 0.95) then
                g_currentMission:showBlinkingWarning(g_i18n:getText("unloadBaleOrSwitchToAuto"));
            end;
        end;
    end;

    if self:getIsActive() and self:getIsTurnedOn() then
        if self.balerStates[self.currentBalerState].updateFunction ~= nil then
            self.balerStates[self.currentBalerState].updateFunction(self, dt);
        end;
        if self.wrapperStates[self.currentWrapperState].updateFunction ~= nil then
            self.wrapperStates[self.currentWrapperState].updateFunction(self, dt);
        end;
        
        if self.balerCurrentBale ~= nil and self.currentBalerState ~= Ultima.STATE_dropBale then
            rotate(self.balerCurrentBale.node, unpack(self.balerBaleConductNodeRot));
        end;
    end;
    
    for a=1, 4 do
        local sideDoor = self.sideDoors[a];
        if sideDoor:getIsOpened() and sideDoor.playerInTrigger then
            if not (g_gui.currentGui ~= nil or g_currentMission.isPlayerFrozen) then
                -- switch net roles and refill main role
                if self.sideDoors.sideDoorFrontLeft == sideDoor and self.manualNetRoleRefill then
                    if self.netRoleTop.length <= 0 then
                        if getVisibility(self.netRoleStorageMiddle.nodeId) then 
                            g_currentMission:addHelpButtonText(g_i18n:getText("refillNetRole"), InputBinding.IMPLEMENT_EXTRA2);
                            if InputBinding.hasEvent(InputBinding.IMPLEMENT_EXTRA2) then
                                if self:getAnimationTime(self.netRoleStorageMiddle.refillAnimName) == 0 then
                                    self:playAnimation(self.netRoleStorageMiddle.refillAnimName, 1, 0);
                                end;
                            end;
                        elseif getVisibility(self.netRoleStorageLeft.nodeId) then
                            g_currentMission:addHelpButtonText(g_i18n:getText("switchNetRolePosition"), InputBinding.IMPLEMENT_EXTRA2);
                            if InputBinding.hasEvent(InputBinding.IMPLEMENT_EXTRA2) then
                                self:setNetRoleOnStoragePlace(1, false);
                                self:setNetRoleOnStoragePlace(2, true);
                            end;
                        elseif getVisibility(self.netRoleStorageRight.nodeId) then    
                            g_currentMission:addHelpButtonText(g_i18n:getText("switchNetRolePosition"), InputBinding.IMPLEMENT_EXTRA2);
                            if InputBinding.hasEvent(InputBinding.IMPLEMENT_EXTRA2) then
                                self:setNetRoleOnStoragePlace(3, false);
                                self:setNetRoleOnStoragePlace(2, true);
                            end;
                        else
                            g_currentMission:showBlinkingWarning(g_i18n:getText("youNeedToRefillNetRoles"));
                            if self:getIsTurnedOn() and self:allowPickingUp() then
                                self.speedLimit = 0;
                            end;
                        end;
                    else
                        local netNeeded = math.max(1, self.baleSizes[self.baleSizeIndex].size * math.pi * self.numNetWraps); -- prevent dividing with 0
                        local numBales = math.floor(self.netRoleTop.length / netNeeded);
                        local text = string.gsub(g_i18n:getText("netRoleFillState"), "percent" , tostring(math.floor(self.netRoleTop.length/self.netRoleDefaultLenght*100)));
                        local text = string.gsub(text, "numBales", tostring(numBales));
                        g_currentMission:addExtraPrintText(text);
                    end;
                    
                    -- refill net roles
                    if g_currentMission.player.currentRoleMount ~= nil and g_currentMission.player.currentRoleType == "net" then
                        local index;
                        if not getVisibility(self.netRoleStorageMiddle.nodeId) then
                            index = 2;
                        elseif not getVisibility(self.netRoleStorageLeft.nodeId) then
                            index = 1;
                        elseif not getVisibility(self.netRoleStorageRight.nodeId) then
                            index = 3;
                        end;
                        if index ~= nil then
                            g_currentMission:addHelpButtonText(g_i18n:getText("insertNetRole"), InputBinding.IMPLEMENT_EXTRA2);
                            if InputBinding.hasEvent(InputBinding.IMPLEMENT_EXTRA2) then
                                g_currentMission.player.currentRoleMount:playerRemoveRole(g_currentMission.player);
                                self:setNetRoleOnStoragePlace(index, true);
                            end;
                        end;
                    end;
                elseif self.manualFoilRoleRefill then    -- foil roles            
                    -- find a full and an empty role
                    local storage = self.foilRoleStorages[sideDoor.animName];
                    local holderFullIndex, holderEmptyIndex;
                    if storage ~= nil then
                        for b=1, table.getn(storage.holders) do
                            if getVisibility(storage.holders[b]) then
                                if holderFullIndex == nil then
                                    holderFullIndex = b;
                                end;
                            else
                                if holderEmptyIndex == nil then
                                    holderEmptyIndex = b;
                                end;
                            end;
                            if holderFullIndex ~= nil and holderEmptyIndex ~= nil then
                                break;
                            end;
                        end;
                    end;
                    
                    if g_currentMission.player.currentRoleMount ~= nil then
                        -- fill holder
                        if holderEmptyIndex ~= nil then
                            if g_currentMission.player.currentRoleType == "foil" then
                                g_currentMission:addHelpButtonText(g_i18n:getText("refillFoilRole"), InputBinding.IMPLEMENT_EXTRA2);
                                if InputBinding.hasEvent(InputBinding.IMPLEMENT_EXTRA2) then
                                    g_currentMission.player.currentRoleMount:playerRemoveRole(g_currentMission.player);
                                    self:foilStorageSetRole(self.foilRoleStorages[sideDoor.animName].index, holderEmptyIndex, true);
                                    break;
                                end;
                            end;
                        end;
                    else
                        -- take role from holder
                        if holderFullIndex ~= nil then
                            g_currentMission:addHelpButtonText(g_i18n:getText("takeFoilRole"), InputBinding.IMPLEMENT_EXTRA2);
                            if InputBinding.hasEvent(InputBinding.IMPLEMENT_EXTRA2) then
                                self:foilStorageSetRole(self.foilRoleStorages[sideDoor.animName].index, holderFullIndex, false);
                                self:playerTakeRole(g_currentMission.player);
                                break;
                            end;
                        end;
                    end;
                end;
            end;
        end;
    end;
    
    if self.manualFoilRoleRefill then
        for a=1, table.getn(self.wrapperFoilHolders) do
            if self.wrapperFoilHolders[a].playerInTrigger then
                if not (g_gui.currentGui ~= nil or g_currentMission.isPlayerFrozen) then
                    -- refill wrapper foil holders
                    if self.wrapperFoilHolders[a].remainingFoilLength <= 0 then
                        if g_currentMission.player.currentRoleMount ~= nil then
                            g_currentMission:addHelpButtonText(g_i18n:getText("refillFoilRole"), InputBinding.IMPLEMENT_EXTRA2);
                            if InputBinding.hasEvent(InputBinding.IMPLEMENT_EXTRA2) then
                                g_currentMission.player.currentRoleMount:playerRemoveRole(g_currentMission.player);
                                self:setWrapFoilLength(a, self.maxFoilLength);
                            end;
                        end;
                    else -- show fill state
                        local foilLength = self.wrapperFoilHolders[a].remainingFoilLength*2;
                        local foilPerBale = self.baseFoilUsePerBale*self.baleSizes[self.baleSizeIndex].size;
                    
                        local numBales = math.ceil(foilLength / foilPerBale);
                        local text = string.gsub(g_i18n:getText("foilHoldersFillState"), "numBales" , tostring(numBales));
                        local text = string.gsub(text, "percent" , tostring(math.floor(foilLength/self.maxFoilLength*100/2)));
                    
                        g_currentMission:addExtraPrintText(text);
                    end;
                end;
            end;
        end;
    end;
    
    if self.isServer and self.manualNetRoleRefill then
        -- trigger chamber role refill
        if self:getAnimationTime(self.netRoleStorageMiddle.refillAnimName) == 1 then
            self:setNetRoleFillState(self.netRoleTop.maxLength);
            self:setNetRoleOnStoragePlace(2, false);
            self:playAnimation(self.netRoleStorageMiddle.refillAnimName, -1, 1);
        end;
    end;
    
    if not (g_gui.currentGui ~= nil or g_currentMission.isPlayerFrozen) then
        -- side door control
        for a=1, self.numSideDoors do
            local sideDoor = self.sideDoors[a];
            if sideDoor.playerInTrigger then
                if g_currentMission.player.isEntered then
                    if sideDoor:getIsClosed() then
                        g_currentMission:addHelpButtonText(g_i18n:getText("openSideDoor"), InputBinding.IMPLEMENT_EXTRA);
                        if InputBinding.hasEvent(InputBinding.IMPLEMENT_EXTRA) then
                            sideDoor:open();
                            SoundUtil.playSample(self.sideDoorOpenSound, 1, 0);
                        end;
                    elseif sideDoor:getIsOpened() then
                        if (sideDoor.animName == "sideDoorFrontLeft" and self:getAnimationTime(self.netRoleStorageMiddle.refillAnimName) == 0) or sideDoor.animName ~= "sideDoorFrontLeft" then
                            g_currentMission:addHelpButtonText(g_i18n:getText("closeSideDoor"), InputBinding.IMPLEMENT_EXTRA);
                            if InputBinding.hasEvent(InputBinding.IMPLEMENT_EXTRA) then
                                sideDoor:close();
                                SoundUtil.playSample(self.sideDoorCloseSound, 1, 0);
                            end;
                        end;
                    end;
                end;
            end;
        end;
        if self.dropmat.playerInTrigger then
            if self.dropmat.isFolding then
                g_currentMission:addHelpButtonText(g_i18n:getText("unfoldDropmat"), InputBinding.IMPLEMENT_EXTRA2);
            else
                g_currentMission:addHelpButtonText(g_i18n:getText("foldDropmat"), InputBinding.IMPLEMENT_EXTRA2);
            end;
            
            if InputBinding.hasEvent(InputBinding.IMPLEMENT_EXTRA2) then
                if self:getAnimationTime(self.dropmat.animName) == 0 then
                    self:playAnimation(self.dropmat.animName, 1, 0);
                    self.dropmat.isFolding = true;
                elseif self:getAnimationTime(self.dropmat.animName) == 1 then
                    self:playAnimation(self.dropmat.animName, -1, 1);
                    self.dropmat.isFolding = false;
                elseif self.animations[self.dropmat.animName].currentSpeed > 0 then
                    self:playAnimation(self.dropmat.animName, -1);
                    self.dropmat.isFolding = false;
                elseif self.animations[self.dropmat.animName].currentSpeed < 0 then
                    self:playAnimation(self.dropmat.animName, 1);
                    self.dropmat.isFolding = true;
                end;
            end;
        end;
    end;
    if self:getAnimationTime(self.dropmat.animName) == 0 then -- and self.dropmat.isFolding then
        local x, y, z = getWorldTranslation(self.dropmat.raycastNode);
        raycastClosest(x, y, z, 0, -1, 0, "dropmatRaycast", 15, self, 6, true);
    end;
    
    if self.isClient then
        Utils.updateRotationNodes(self, self.ultimaTurnOnRotNodes, dt, self:getIsActive() and self:getIsTurnedOn());
        Utils.updateScrollers(self.UVScrollParts, dt, self:getIsActive() and self:getIsTurnedOn());
    end;
end;

function Ultima:dropmatRaycast(hitObjectId, x, y, z, distance, nx, ny, nz, subShapeIndex)
    local lx, ly, lz = worldToLocal(getParent(self.dropmat.jointNode), x, y, z);
    local height = math.min(self.dropmat.maxJointY, math.max(self.dropmat.minJointY, ly));
    local jx, jy, jz = getTranslation(self.dropmat.jointNode);
    setTranslation(self.dropmat.jointNode, jx, height, jz);
end;

function Ultima:updateTick(dt)
    if self.lastTrackingPos[1] == 0 and self.lastTrackingPos[2] == 0 and self.lastTrackingPos[3] == 0 then
        self.lastTrackingPos = {getWorldTranslation(self.components[1].node)}
    end;
    
    if self.hud.baseElement.visibility ~= self:getIsActiveForInput(false) then
        if self:getIsActiveForInput(false) then
            self.hud:open();
        else
            self.hud:close();
        end;
    end;
    
    if self:getIsTurnedOn() then
        if not SoundUtil.isSamplePlaying(self.novoGripStartSound, 1.8*dt) then
            if not SoundUtil.isSamplePlaying(self.novoGrip3DSound, 1.8*dt) then
                SoundUtil.play3DSample(self.novoGrip3DSound);
            end;
        end;
    end;

    if self:getIsTurnedOn() and self:allowPickingUp() then 
        if self.isServer then
            local workAreas, fieldNotOwned, areaFound = self:getTypedNetworkAreas(WorkArea.AREATYPE_BALER, false);
            local area, fillType = Baler.processBalerAreas(self, workAreas, self.balerFillTypes);
            
            if area > 0 and self:allowFillType(fillType) then
                local literPerPx = FruitUtil.getFillTypeLiterPerSqm(fillType, 1) * g_currentMission:getFruitPixelsToSqm();
                local amount = area * literPerPx * 1; -- maybee add a compression factor for each fillType?
                local freeSpace = self:getUnitCapacity(self.preChamberUnitIndex)-self:getUnitFillLevel(self.preChamberUnitIndex);
            
                self:setUnitFillLevel(self.preChamberUnitIndex, self:getUnitFillLevel(self.preChamberUnitIndex)+math.min(amount, freeSpace), fillType, true);
                            
                -- calculate average input per meter
                local x, y, z = getWorldTranslation(self.components[1].node);
                local tp = self.lastTrackingPos;
                local dist = Utils.vector3Length(tp[1]-x, tp[2]-y, tp[3]-z);
                self.lastTrackingPos = {getWorldTranslation(self.components[1].node)}
                table.insert(self.inputTracker, {distance=dist, input=amount});
                
                local fullDistance = 0;
                local fullInput = 0;
                local numRemoved = 0;
                
                -- !!! queue running backwards through table
                for a=table.getn(self.inputTracker), 1, -1 do
                    if fullDistance < self.trackDistance then
                        fullDistance = fullDistance + self.inputTracker[a-numRemoved].distance;
                        fullInput = fullInput + self.inputTracker[a-numRemoved].input;
                    else
                        table.remove(self.inputTracker, a-numRemoved);
                        numRemoved = numRemoved + 1;
                    end;
                end;
                self.averageInputPerM = fullInput/fullDistance;
                
                if amount > 0 then
                    self.pickupEffectActiveTime = 1000;
                end;
            end;
            
            local capacity, fillLevel;
            if self.currentBalerState == Ultima.STATE_usualLoading or self.currentBalerState == Ultima.STATE_passLoad then
                capacity = self:getUnitCapacity(self.mainChamberUnitIndex);
                fillLevel = self:getUnitFillLevel(self.mainChamberUnitIndex);
                if self.currentBaleDropMode == Ultima.DROPMODE_COLLECT and self:getUnitFillLevel(self.mainChamberUnitIndex) > (self:getUnitCapacity(self.mainChamberUnitIndex)*0.95) then
                    capacity = capacity + self:getUnitCapacity(self.mainChamberUnitIndex) + self:getUnitCapacity(self.preChamberUnitIndex);
                    fillLevel = fillLevel + self:getUnitFillLevel(self.preChamberUnitIndex);
                end;
            else
                capacity = self:getUnitCapacity(self.preChamberUnitIndex);
                fillLevel = self:getUnitFillLevel(self.preChamberUnitIndex);
            end;
            
            local speedLimit = self.maxSpeedLimit;
            local timeRemaining = math.max(self:getRemainingAnimationTime(),1); -- prevent us from dividing with 0 
            local distanceToCover = (capacity-fillLevel)/self.averageInputPerM; 
            speedLimit = (distanceToCover/(timeRemaining/1000))*3.6;
            speedLimit = math.min(speedLimit, self.maxSpeedLimit);
            
            -- buffer speed limit to drive more smoove
            self.speedLimitBufferNum = self.speedLimitBufferNum + 1;
            self.speedLimitBufferAmount = self.speedLimitBufferAmount + speedLimit;
            
            self.speedLimit = self.speedLimitBufferAmount / self.speedLimitBufferNum;
            if self.speedLimitBufferNum > 50 then
                self.speedLimitBufferNum = 0;
                self.speedLimitBufferAmount = 0;
            end;
            
            -- stop baler in case that the wrapper bale has not been dropped
            if self.currentBaleDropMode == Ultima.DROPMODE_MANUAL then
                if self.wrapperCurrentBale ~= nil and self:getUnitFillLevel(self.mainChamberUnitIndex) > (self:getUnitCapacity(self.mainChamberUnitIndex) * 0.95) then
                    self.speedLimit = 0;
                end;
            end;
        end;
    else
        self.speedLimit = math.huge;
    end;
    
    if self.pickupEffect ~= nil then
        if self.isServer then
            self.pickupEffectActiveTime = math.max(0, self.pickupEffectActiveTime - dt);
            if self.pickupEffectActiveTime <= 0 then
                self.pickupEffectActive = false;
            else
                self.pickupEffectActive = true;
            end;
        end;
        
        if self.pickupEffectActive then
            local fillType = self:getUnitFillType(self.mainChamberUnitIndex);
            if fillType == FillUtil.FILLTYPE_UNKNOWN then
                fillType = self:getUnitFillType(self.preChamberUnitIndex);
            end;
            EffectManager:setFillType(self.pickupEffect, fillType);
            EffectManager:startEffects(self.pickupEffect);
        else
            EffectManager:stopEffects(self.pickupEffect);
        end;
    end;
        
    self.netBindingActiveImg:setVisibility(self.currentBalerState == Ultima.STATE_netBinding);
    self.netBindingPassiveImg:setVisibility(self.currentBalerState ~= Ultima.STATE_netBinding);
    
    self.ppcOpenImg:setVisibility(self.currentBalerState == Ultima.STATE_netBinding or self.currentBalerState == Ultima.STATE_dropBale);
    self.ppcClosedImg:setVisibility(self.currentBalerState == Ultima.STATE_passLoad or self.currentBalerState == Ultima.STATE_usualLoading);
    
    self.baleDoorOpeningImg:setVisibility(self.currentBalerState == Ultima.STATE_dropBale);
    self.baleDoorClosedImg:setVisibility(self.currentBalerState ~= Ultima.STATE_dropBale);
    
    self.wrapperActiveImg:setVisibility(self.currentWrapperState == Ultima.WRAPPERSTATE_wrapping);
    self.wrapperPassiveImg:setVisibility(self.currentWrapperState ~= Ultima.WRAPPERSTATE_wrapping);
end;

function floor2(value)
    return tostring(math.floor(value*100));
end;

function Ultima:draw()
    if self:getIsActiveForInput() then
        if self.baleCountDay > 0 then
            g_currentMission:addHelpButtonText(g_i18n:getText("resetBaleCounter"), InputBinding.IMPLEMENT_EXTRA4); 
        end;
        if self.hud.hideGUI then
            g_currentMission:addHelpButtonText(g_i18n:getText("showGUI"), InputBinding.HIDEGUI); 
        else
            g_currentMission:addHelpButtonText(g_i18n:getText("hideGUI"), InputBinding.HIDEGUI); 
        end;

        if self:getIsTurnedOn() then
            local text = string.gsub(g_i18n:getText("setBaleSize"), "sizeX", tostring(self.baleSizes[self.baleSizeIndex].size));
            g_currentMission:addHelpButtonText(text, InputBinding.SETBALESIZE); 
            if self.wrapperIsActive then
                g_currentMission:addHelpButtonText(g_i18n:getText("deactivateWrapper"), InputBinding.SETWRAPPERACTIVE); 
            else
                g_currentMission:addHelpButtonText(g_i18n:getText("activateWrapper"), InputBinding.SETWRAPPERACTIVE); 
            end;
            
            local mode;
            if self.currentBaleDropMode == Ultima.DROPMODE_AUTO then
                mode = g_i18n:getText("mode_auto");
            elseif self.currentBaleDropMode == Ultima.DROPMODE_MANUAL then
                mode = g_i18n:getText("mode_manual");
            elseif self.currentBaleDropMode == Ultima.DROPMODE_COLLECT then
                mode = g_i18n:getText("mode_collect");
            end;
            local text = string.gsub(g_i18n:getText("setBaleDropMode"), "modeX", mode);
            g_currentMission:addHelpButtonText(text, InputBinding.SETBALEDROPMODE); 
        end;
    end;
    if self:getIsActiveForInput(false) and self:getIsTurnedOn() then
        if self.currentBaleDropMode == Ultima.DROPMODE_MANUAL then
            if self.currentWrapperState == Ultima.WRAPPERSTATE_dropCheck then
                g_currentMission:addHelpButtonText(g_i18n:getText("dropBale"), InputBinding.WRAPPERDROPBALE); 
            end;
        end;
    end;
end;

function Ultima:onDeactivateSounds()
    if self.isClient then
        SoundUtil.stopSample(self.balerAlarmSound, true);
    end;
end;

function Ultima:foilStorageSetRole(storageIndex, holderIndex, visibility, noEventSend)
    setVisibility(self.foilRoleStorages[storageIndex].holders[holderIndex], visibility);

    if not noEventSend then
        FoilStorageSetRoleEvent.sendEvent(self, storageIndex, holderIndex, visibility);
    end;
end;

function Ultima:setBalerState(state, noEventSend)
    self.currentBalerState = state;
    if self.balerStates[state].startFunction ~= nil then
        self.balerStates[state].startFunction(self);
    end;

    if not noEventSend then
        SetBalerStateEvent.sendEvent(self, state);
    end;
end;

function Ultima:setWrapperState(state, noEventSend)
    self.currentWrapperState = state;
    if self.wrapperStates[state].startFunction ~= nil then
        self.wrapperStates[state].startFunction(self);
    end;
    if not noEventSend then
        SetWrapperStateEvent.sendEvent(self, state);
    end;
end;

function Ultima:getMPCFillType()
    return self:getUnitFillType(self.mainChamberUnitIndex);
end;

function Ultima:getPPCFillType()
    return self:getUnitFillType(self.preChamberUnitIndex);
end;

function Ultima:balerCreateBale(baleSizeIndex, fillType, linkToWrapper) -- creates temporary bale used for animations
    if self.baleSizes[baleSizeIndex].fillTypes[fillType] == nil then
    end;
        
    local bale = {
        sizeIndex = baleSizeIndex,
        fillType = fillType,
        fillLevel = self.baleSizes[baleSizeIndex].fillLevel,
        willBeWrapped = self.wrapperIsActive and self.baleSizes[baleSizeIndex].fillTypes[fillType].isWrapAble;
    }
    
    local fileName = (bale.willBeWrapped and self.baleSizes[baleSizeIndex].fillTypes[fillType].fileNameForWrapping ~= nil and self.baleSizes[baleSizeIndex].fillTypes[fillType].fileNameForWrapping or self.baleSizes[baleSizeIndex].fillTypes[fillType].fileName);    
    local i3dNode = Utils.loadSharedI3DFile(fileName, self.baseDirectory, false, false, false);    
        
    if fileName ~= nil and i3dNode ~= 0 then
        bale.fileName = fileName;
        bale.node = getChildAt(i3dNode, 0);
        
        unlink(bale.node);
        delete(i3dNode);
        
        setRigidBodyType(bale.node, "NoRigidBody");
        local linkNode = self.balerBaleConductNode;
        if linkToWrapper ~= nil and linkToWrapper then
            linkNode = self.wrapperBaleConductNode;
        end
        
        link(linkNode, bale.node);
    end;
    
    -- fix, since chamber fill anim was made for 1,5er bales only
    local scale = 1.5/self.baleSizes[self.baleSizeIndex].size;
    setScale(self.balerBaleConductNode, scale, scale, 1);
    
    return bale;
end;

function Ultima:setBaleSize(sizeIndex, noEventSend)
    if self.baleSizes[sizeIndex] ~= nil then
        self.baleSizeIndex = sizeIndex;
    else
        self.baleSizeIndex = 1;
    end;
    
    if not noEventSend then
        SetBaleSizeEvent.sendEvent(self, self.baleSizeIndex);
    end;
end;

function Ultima:setWrapperActive(isActive, noEventSend)
    self.wrapperIsActive = isActive;
    if not noEventSend then
        SetWrapperActiveEvent.sendEvent(self, isActive);
    end;
end;

function Ultima:setBaleDropMode(mode, noEventSend)
    if mode <= 3 then
        self.currentBaleDropMode = mode;
    else
        self.currentBaleDropMode = 1;
    end;
    if not noEventSend then
        SetBaleDropModeEvent.sendEvent(self, self.currentBaleDropMode);
    end;
end;

function Ultima:allowPickingUp(oldFunc)
    return oldFunc(self);
end;

function Ultima:getRemainingAnimationTime()
    local animTime = 0;

    -- pre press chamber
    if self.currentBalerState == Ultima.STATE_netBinding or self.currentBalerState == Ultima.STATE_dropBale then
        if self.currentBalerState == Ultima.STATE_netBinding then
            animTime = animTime + self.netBindingTimer;
        end;
        if self.currentBalerState <= Ultima.STATE_dropBale then
            local baleSizeIndex;
            if self.balerCurrentBale ~= nil then
                baleSizeIndex = self.balerCurrentBale.sizeIndex;
            else
                baleSizeIndex = self.wrapperCurrentBale.sizeIndex;
            end;
        
            if self.currentBalerState == Ultima.STATE_netBinding then
                animTime = animTime + self.baleSizes[baleSizeIndex].dropBaleAnimation.duration; -- anim stops at 100% 
            else
                animTime = animTime + self.baleSizes[baleSizeIndex].dropBaleAnimation.duration - self.baleSizes[baleSizeIndex].dropBaleAnimation:getTime();
            end;
        end;
        if animTime <= 0 then
            return 1;
        else
            return animTime+self.remainingAnimTimeBuffer;
        end;
    end;
    
    -- main press chamber
    if self.currentBalerState == Ultima.STATE_usualLoading or self.currentBalerState == Ultima.STATE_passLoad then
        local baleSize;
        local supportsWrapping;
        if self.wrapperCurrentBale ~= nil then
            baleSize = self.baleSizes[self.wrapperCurrentBale.sizeIndex];
            fillType = self.wrapperCurrentBale.fillType;
        else
            return 1; -- no bale yet
        end;

        if self.currentWrapperState == Ultima.WRAPPERSTATE_fetchBale then
            animTime = animTime + self:getRealAnimationTime(baleSize.wrapperReceiveAnim);
        end;    

        -- wrapper anim
        if self.wrapperCurrentBale.willBeWrapped then
            -- note: animation stops at the end and is not set to 0ms until anim starts
            if self.currentWrapperState < Ultima.WRAPPERSTATE_wrapping then
                animTime = animTime + baleSize.wrapAnimation.duration; 
            elseif self.currentWrapperState == Ultima.WRAPPERSTATE_wrapping then
                animTime = animTime + baleSize.wrapAnimation.duration - baleSize.wrapAnimation:getTime();
            end;
        end;        
    
        -- drop anim
        local currentSpeed = self.animations[baleSize.wrapperDropAnimName].currentSpeed;
        if self.currentWrapperState < Ultima.WRAPPERSTATE_dropBale then
            animTime = animTime + (self:getAnimationDuration(baleSize.wrapperDropAnimName)*2);
        end;
        if self.currentWrapperState == Ultima.WRAPPERSTATE_dropBale then
            if currentSpeed > 0 then -- still not dropped
                animTime = animTime + self:getAnimationDuration(baleSize.wrapperDropAnimName) - self:getRealAnimationTime(baleSize.wrapperDropAnimName);
            else -- already dropped, animation running backwards
                --animTime = animTime + self:getAnimationDuration(baleSize.wrapperDropAnimName) - self:getRealAnimationTime(baleSize.wrapperDropAnimName);
            end;
        end;
        if animTime <= 0 then
            return 1;
        else
            return animTime+self.remainingAnimTimeBuffer;
        end;
    end;
end;

function Ultima:playerTakeRole(player, noEventSend)
    if player ~= nil and player.currentRoleMount == nil then
        local nodeId = clone(self.foilRoleCloneNode, true, false, false);
        link(player.toolsRootNode, nodeId);
        setVisibility(nodeId, true);
        player.currentRoleMount = self;
        player.currentRoleType = "foil";
        player.currentRoleNodeId = nodeId;
        setTranslation(nodeId, unpack(self.playerRolePos));
        setRotation(nodeId, unpack(self.playerRoleRot));
        
        if not noEventSend then
            PlayerTakeRoleEvent.sendEvent(self, player, true);
        end;
    end;
end;

function Ultima:playerRemoveRole(player, noEventSend)
    if player ~= nil and player.currentRoleMount ~= nil then
        unlink(player.currentRoleNodeId);
        player.currentRoleMount = nil;
        player.currentRoleType = nil;
        delete(player.currentRoleNodeId);
        player.currentRoleNodeId = nil;
    end;

    if not noEventSend then
        PlayerTakeRoleEvent.sendEvent(self, player, false);
    end;
end;

function Ultima:setNetRoleOnStoragePlace(placeIndex, visibility, noEventSend)
    local storage = self.netRoleStorages[placeIndex];
    setVisibility(storage.nodeId, visibility);
    if visibility then
        storage.length = self.netRoleDefaultLenght;
    end;
    
    if not noEventSend then
        SetNetRoleOnStorageEvent.sendEvent(self, placeIndex, visibility); 
    end;
end;

function Ultima:setNetRoleFillState(length) -- only server to client
    if self.manualNetRoleRefill then
        self.netRoleTop.length = math.max(0, length);
        local animTime = 1-math.cos((math.pi/2)-(math.pi/2)*(self.netRoleTop.length/self.netRoleTop.maxLength));
        self:setAnimationTime(self.netRoleTop.fillAnim, animTime);
        if self.isServer then
            g_server:broadcastEvent(SetNetRoleFillStateEvent:new(self, length), nil, nil, self);
        end;
    end;
end;

function Ultima:usualLoadingUpdate(dt)
    if self.isServer then
        self:passLoadToMainChamber(dt);
        if self:getUnitFillLevel(self.mainChamberUnitIndex) >= self:getUnitCapacity(self.mainChamberUnitIndex) then
            if self.currentWrapperState == Ultima.WRAPPERSTATE_empty then
                self:setBalerState(Ultima.STATE_netBinding);
            end;
        end; 
    end;
end;

function Ultima:netBindingStart()
    if self:getIsActiveForSound() then
        SoundUtil.playSample(self.netBindingStartSound, 1, 0);
    end;
    self.netBindingTimer = self.netBindingTimerTime;
    self:setWrapperState(Ultima.WRAPPERSTATE_tiltTable);
    self.chamberFillAnim:stop();
    if self.isServer then
        local netNeeded = self.baleSizes[self.balerCurrentBale.sizeIndex].size * math.pi * self.numNetWraps;
        self:setNetRoleFillState(self.netRoleTop.length-netNeeded);
    end;
end;

function Ultima:netBindingUpdate(dt)
    if self.netRoleTop.length > 0 then
        rotate(self.netRoleTop.role, unpack(self.netRoleTop.netRoleRotSpeed));
        self.netBindingTimer = self.netBindingTimer - dt;
        if self.netBindingTimer <= 0 and self.isServer then
            self:setBalerState(Ultima.STATE_dropBale);
        end;
    else
        if self:getIsActiveForInput(false) then
            g_currentMission:showBlinkingWarning(g_i18n:getText("netRoleEmpty"));
            if self:getIsTurnedOn() and self:allowPickingUp() then
                self.speedLimit = 0;
            end;
        end;
    end;
end;

function Ultima:dropBaleStart()
    if self:getIsActiveForSound() then
        SoundUtil.playSample(self.netBindingStopSound, 1, 0);
        SoundUtil.playSample(self.baleDoorOpenSound, 1, 0);
    end;
    self.baleSizes[self.balerCurrentBale.sizeIndex].dropBaleAnimation:play();
end;

function Ultima:dropBaleUpdate()
    if self.isServer and self.balerCurrentBale ~= nil then
        if self.baleSizes[self.balerCurrentBale.sizeIndex].dropBaleAnimation:getTime() > self.baleSizes[self.balerCurrentBale.sizeIndex].receiveBaleAnimTime then
            if self.currentWrapperState == Ultima.WRAPPERSTATE_tiltTable then    
                self:linkBaleToWrapper();
                self:setWrapperState(Ultima.WRAPPERSTATE_fetchBale);
                self:setUnitFillLevel(self.mainChamberUnitIndex, 0, FillUtil.FILLTYPE_UNKNOWN, true);
            end;
        end;
    end;
    -- note: balerCurrentBale becomes wrapperCurrentBale !!
    if self.wrapperCurrentBale ~= nil then
        if self.baleSizes[self.wrapperCurrentBale.sizeIndex].dropBaleAnimation:getTime() >= self.baleSizes[self.wrapperCurrentBale.sizeIndex].dropBaleAnimation.duration then
            if self.isServer then
                self:setBalerState(Ultima.STATE_passLoad);
            end;
        end;
    end;
end;

function Ultima:linkBaleToWrapper() -- server to client
    -- hier wird der temporre Ballen von der Presse an den Wickler bergeben
    unlink(self.balerCurrentBale.node);
    link(self.wrapperBaleConductNode, self.balerCurrentBale.node);
    self.wrapperCurrentBale = self.balerCurrentBale;
    self.balerCurrentBale = nil;

    if self.isServer then
        g_server:broadcastEvent(LinkBaleToWrapperEvent:new(self), nil, nil, self);
    end;
end;

function Ultima:passLoadStart()
    if self:getIsActiveForSound() then
        SoundUtil.playSample(self.baleDoorCloseSound, 1, 0, nil);
    end;
    self.baleSizes[self.wrapperCurrentBale.sizeIndex].dropBaleAnimation:stop();
end;

function Ultima:passLoadUpdate(dt)
    if self.isServer then
        self:passLoadToMainChamber(dt);
        if self:getUnitFillLevel(self.preChamberUnitIndex) <= 0 or self:getUnitFillLevel(self.mainChamberUnitIndex) >= self:getUnitCapacity(self.mainChamberUnitIndex) then
            self:setBalerState(Ultima.STATE_usualLoading);
        end;
    end;
end;

function Ultima:passLoadToMainChamber(dt)
    if self.isServer and self:getUnitFillLevel(self.preChamberUnitIndex) > 0 then
        local amount = math.min(self:getUnitFillLevel(self.preChamberUnitIndex), self.ppcPassThrough/1000*dt);
        local freeSpace = self:getUnitCapacity(self.mainChamberUnitIndex)-self:getUnitFillLevel(self.mainChamberUnitIndex);
        local delta = math.min(amount, freeSpace);
        
        local fillType = self:getUnitFillType(self.preChamberUnitIndex);
    
        self:setUnitFillLevel(self.preChamberUnitIndex, self:getUnitFillLevel(self.preChamberUnitIndex)-delta, fillType, true);
        self:setUnitFillLevel(self.mainChamberUnitIndex, self:getUnitFillLevel(self.mainChamberUnitIndex)+delta, fillType, true);
    end;
end;

function Ultima:wrapperTiltTableStart() -- !! XML Animation !!
    if self.isServer then -- anim wird vom animatedVehicle gesyncht
        self:playAnimation(self.baleSizes[self.balerCurrentBale.sizeIndex].wrapperReceiveAnim, 1, 0, false);
    end;
end;

function Ultima:wrapperTiltTableUpdate(dt)
    if self:getIsActiveForSound() then
        if self:getAnimationTime(self.baleSizes[self.balerCurrentBale.sizeIndex].wrapperReceiveAnim) > 0.95 and self:getAnimationTime(self.baleSizes[self.balerCurrentBale.sizeIndex].wrapperReceiveAnim) < 1 then
            if not SoundUtil.isSamplePlaying(self.wrapperDropBaleSound, 1.8*dt) then
                SoundUtil.playSample(self.wrapperDropBaleSound, 1, 0);
            end;
        end;
    end;
end;

function Ultima:wrapperFetchBaleStart() -- !! XML Animation !!
    if self.isServer then -- anim wird vom animatedVehicle gesyncht
        self:playAnimation(self.baleSizes[self.wrapperCurrentBale.sizeIndex].wrapperReceiveAnim, -1, 1, false);
    end;
end;

function Ultima:wrapperFetchBaleUpdate() -- !! XML Animation !!
    if self:getAnimationTime(self.baleSizes[self.wrapperCurrentBale.sizeIndex].wrapperReceiveAnim) == 0 then
        if self.wrapperCurrentBale.willBeWrapped then
            if self.wrapperFoilHolders[1].remainingFoilLength > 0 and self.wrapperFoilHolders[2].remainingFoilLength > 0 then
                if self.isServer then
                    self:setWrapperState(Ultima.WRAPPERSTATE_wrapping);
                end;
            elseif self:getIsActiveForInput(false) then
                g_currentMission:showBlinkingWarning(g_i18n:getText("foilRoleEmpty"));
            end;
        elseif self.isServer then
            self:wrapperCreateBale(self.wrapperCurrentBale.sizeIndex, self.wrapperCurrentBale.fillType, self.wrapperCurrentBale.fileName, self.wrapperCurrentBale.node);
            self:setWrapperState(Ultima.WRAPPERSTATE_dropCheck);
        end;
    end;
end;

function Ultima:wrapperWrappingStart()
    self.baleSizes[self.wrapperCurrentBale.sizeIndex].wrapAnimation:play();
    self.wrapperRolesAnim:play();
    if self:getIsActiveForSound() then
        SoundUtil.playSample(self.wrappingStartSound, 1, 0);
    end;
    for a=1, table.getn(self.foilWrapCutMesh) do
        setVisibility(self.foilWrapCutMesh[a], false);
    end;
    for a=1, table.getn(self.foilWrapMesh) do
        setVisibility(self.foilWrapMesh[a], true);
    end;
end;

function Ultima:wrapperWrappingUpdate(dt)
    if self.baleSizes[self.wrapperCurrentBale.sizeIndex].wrapAnimation:getTime() < 19500 then
        if not SoundUtil.isSamplePlaying(self.wrappingStartSound, 1.8*dt) then
            if not SoundUtil.isSamplePlaying(self.wrapping3DSound, 1.8*dt) and self:getIsActiveForSound() then
                SoundUtil.play3DSample(self.wrapping3DSound);
            end;
        end;   --- This is where the dog is burried
    elseif self.baleSizes[self.wrapperCurrentBale.sizeIndex].wrapAnimation:getTime() < 20500 then -- litte dirty but hey...
        SoundUtil.stop3DSample(self.wrapping3DSound);
        if self:getIsActiveForSound() and not SoundUtil.isSamplePlaying(self.wrappingStopSound, 1.8*dt) then
            SoundUtil.playSample(self.wrappingStopSound, 1, 0);
        end;
    end;

    if self.isServer then
        local sizeIndex = self.wrapperCurrentBale.sizeIndex;
        
        if self.baleSizes[sizeIndex].wrapAnimation:getTime() >= self.baleSizes[sizeIndex].createPhysicalBaleTime then 
            if self.wrapperCurrentBale.baleObject == nil then
                local fileName = self.baleSizes[sizeIndex].fillTypes[self.wrapperCurrentBale.fillType].wrappedFileName;
                self:wrapperCreateBale(sizeIndex, self.wrapperCurrentBale.fillType, fileName);
            end;
        end;
        if self.baleSizes[sizeIndex].wrapAnimation:getTime() >= self.baleSizes[sizeIndex].wrapAnimation.duration then
            self:setWrapperState(Ultima.WRAPPERSTATE_dropCheck);
            if self.isServer and self.manualFoilRoleRefill then
                self:setWrapFoilLength(1, self.wrapperFoilHolders[1].remainingFoilLength-(self.baseFoilUsePerBale*self.baleSizes[sizeIndex].size)/2);
                self:setWrapFoilLength(2, self.wrapperFoilHolders[2].remainingFoilLength-(self.baseFoilUsePerBale*self.baleSizes[sizeIndex].size)/2);
            end;
        end;
        if self.wrapperCurrentBale.baleObject ~= nil then
            setJointFrame(self.wrapperCurrentBale.baleJointIndex, 0, self.wrapperBaleConductNode);
        end;
    end;
end;

function Ultima:setWrapFoilLength(index, length, noEventSend) 
    local foilHolder = self.wrapperFoilHolders[index];
    foilHolder.remainingFoilLength = math.max(0, math.min(self.maxFoilLength, length));
    self:setAnimationTime(foilHolder.emptyAnimName, 1-math.cos((math.pi/2)-(math.pi/2)*(foilHolder.remainingFoilLength/self.maxFoilLength)));
    -- Freds solution: 1-cos[(pi/2)-(pi/2)*(akt/ges)]
    if not noEventSend then
        SetWrapFoilLengthEvent.sendEvent(self, index, length);
    end;
end;

-- currently called only at server
-- create the bale that is actually dropped
-- new bale object bale is only on server known to the baler, client does not have any access
function Ultima:wrapperCreateBale(sizeIndex, firstFillType, baleI3dFilename, baleNode) 
    local fillLevel = self.baleSizes[sizeIndex].fillLevel;
    
    if baleNode == nil then
        local i3dNode = Utils.loadSharedI3DFile(baleI3dFilename, self.baseDirectory, false, false, false)
        if i3dNode ~= 0 then
            baleNode = getChildAt(i3dNode, 0);
            unlink(baleNode);
            delete(i3dNode);
        end;
    end;
    
    self:wrapperDeleteOldBale(baleNode);
    
    if self.isServer then        
        local baleObject = Bale:new(self.isServer, self.isClient);
        baleObject.i3dFilename = Utils.getFilename(baleI3dFilename, self.baseDirectory);
        baleObject:setNodeId(baleNode);
        baleObject:register();
        baleObject.baleI3dFilename = baleI3dFilename;
        baleObject.fillLevel = fillLevel;
        local x, y, z = getWorldTranslation(self.wrapperBaleConductNode);
        local rx, ry, rz = getWorldRotation(self.wrapperBaleConductNode);
        link(getRootNode(), baleObject.nodeId);
        setRigidBodyType(baleObject.nodeId, "Dynamic");
        setTranslation(baleObject.nodeId, x, y, z);
        setRotation(baleObject.nodeId, rx, ry, rz);
        
        local constr = JointConstructor:new();
        constr:setActors(self.components[1].node, baleObject.nodeId);
        constr:setJointTransforms(self.wrapperBaleConductNode, self.wrapperBaleConductNode);
                
        for i=1, 3 do
            constr:setRotationLimit(i-1, 0, 0);
            constr:setTranslationLimit(i-1, true, 0, 0);
        end;
        
        constr:setEnableCollision(true);
        self.wrapperCurrentBale.baleJointIndex = constr:finalize();
        self.wrapperCurrentBale.baleObject = baleObject;
        g_currentMission:removeItemToSave(baleObject);        
    end;
end;

function Ultima:wrapperDropBale()
    if self.isServer then
        removeJoint(self.wrapperCurrentBale.baleJointIndex);
        g_currentMission:addItemToSave(self.wrapperCurrentBale.baleObject);
        self.wrapperCurrentBale.baleJointIndex = nil;
        self.wrapperCurrentBale.sizeIndex = nil;
        self.wrapperCurrentBale.fillType = nil;
        self.wrapperCurrentBale = nil;
        self:setBaleCounter(self.baleCountDay+1, self.baleCountLifetime+1);
        g_currentMission.missionStats:updateStats("baleCount", 1);
    end;
end;

function Ultima:wrapperDeleteOldBale(baleNode)
    if self.wrapperCurrentBale ~= nil then
        if not self.isServer or self.wrapperCurrentBale.willBeWrapped then
            delete(self.wrapperCurrentBale.node);
        end;
        
        if self.isServer then
            g_server:broadcastEvent(WrapperDeleteOldBaleEvent:new(self), nil, nil, self);
        end;
    end;
end;

function Ultima:wrapperDropCheckStart()
    if self.baleSizes[self.wrapperCurrentBale.sizeIndex].wrapAnimation:getIsPlaying() then
        self.baleSizes[self.wrapperCurrentBale.sizeIndex].wrapAnimation:stop(); 
    end;
    for a=1, table.getn(self.foilWrapMesh) do
        setVisibility(self.foilWrapMesh[a], false);
    end;
    if self.wrapperCurrentBale.willBeWrapped then
        for a=1, table.getn(self.baleSizes[self.wrapperCurrentBale.sizeIndex].wrapFoilCuts) do
            setVisibility(self.baleSizes[self.wrapperCurrentBale.sizeIndex].wrapFoilCuts[a], true);
        end;
    end;
end;

function Ultima:wrapperDropCheckUpdate()
    if self.currentBaleDropMode == Ultima.DROPMODE_COLLECT then
        -- alarm sound
        if self:getIsActiveForSound() then
            if (self:getUnitFillLevel(self.mainChamberUnitIndex) / self:getUnitCapacity(self.mainChamberUnitIndex)) > 0.96 then
                SoundUtil.playSample(self.balerAlarmSound, 0, 0, nil);
            end;
        end;
    end;
    
    if self.isServer then 
        if self.currentBaleDropMode == Ultima.DROPMODE_AUTO then
            self:setWrapperState(Ultima.WRAPPERSTATE_dropBale);
        elseif self.dropNextBaleDirectly then
            self:setWrapperState(Ultima.WRAPPERSTATE_dropBale);
            self.dropNextBaleDirectly = false;
        elseif self.currentBaleDropMode == Ultima.DROPMODE_COLLECT then
            if self.balerCurrentBale ~= nil and self:getUnitFillLevel(self.mainChamberUnitIndex) >= self.balerCurrentBale.fillLevel then
                self:setWrapperState(Ultima.WRAPPERSTATE_dropBale);
                self.dropNextBaleDirectly = true;
            end;
        end;
    end;
    
    if self.currentBaleDropMode == Ultima.DROPMODE_MANUAL then
        -- alarm sound
        if (self:getUnitFillLevel(self.mainChamberUnitIndex) / self:getUnitCapacity(self.mainChamberUnitIndex)) > 0.87 and self:getIsActiveForSound() then
            SoundUtil.playSample(self.balerAlarmSound, 0, 0, nil);
        end;
        
        if self:getIsActiveForInput(false) then
            if InputBinding.hasEvent(InputBinding.WRAPPERDROPBALE) then
                self:setWrapperState(Ultima.WRAPPERSTATE_dropBale);
            end;
        end;
    end;
end;

function Ultima:wrapperDropBaleStart()
    for a=1, table.getn(self.baleSizes[self.wrapperCurrentBale.sizeIndex].wrapFoilCuts) do
        setVisibility(self.baleSizes[self.wrapperCurrentBale.sizeIndex].wrapFoilCuts[a], false);
    end;
    for a=1, table.getn(self.foilWrapCutMesh) do
        setVisibility(self.foilWrapCutMesh[a], true);
    end;
    self.wrapperBaleSizeIndex = self.wrapperCurrentBale.sizeIndex;
    self:playAnimation(self.baleSizes[self.wrapperCurrentBale.sizeIndex].wrapperDropAnimName, 1, 0, false);
    SoundUtil.stopSample(self.balerAlarmSound);
end;

function Ultima:wrapperDropBaleUpdate()
    if self.isServer then -- update physical bale
        if self.wrapperCurrentBale ~= nil and self.wrapperCurrentBale.baleObject ~= nil then
            setJointFrame(self.wrapperCurrentBale.baleJointIndex, 0, self.wrapperBaleConductNode);
        end;
    end;
    
    if self:getAnimationTime(self.baleSizes[self.wrapperBaleSizeIndex].wrapperDropAnimName) == 1 then
        if self.isServer then
            self:wrapperDropBale(); 
            self:playAnimation(self.baleSizes[self.wrapperBaleSizeIndex].wrapperDropAnimName, -1, 1, false);
        end;
        if self:getIsActiveForSound() then
            SoundUtil.playSample(self.wrapperDropBaleSound, 1, 0);
        end;
    end;
    
    if self:getAnimationTime(self.baleSizes[self.wrapperBaleSizeIndex].wrapperDropAnimName) == 0 and not self:getIsAnimationPlaying(self.baleSizes[self.wrapperBaleSizeIndex].wrapperDropAnimName) then
        self:setWrapperState(Ultima.WRAPPERSTATE_empty);
    end;
end;

function Ultima:onTurnedOn(noEventSend)
    if self:getIsActiveForSound() then
        SoundUtil.playSample(self.novoGripStartSound, 1, 0);
    end;
end;

function Ultima:onTurnedOff(noEventSend)
    self.pickupEffectActive = false;

    SoundUtil.stop3DSample(self.novoGrip3DSound);
    SoundUtil.stop3DSample(self.wrapping3DSound);
    if self:getIsActiveForSound() then
        SoundUtil.playSample(self.novoGripStopSound, 1, 0);
    end;
end;

function Ultima:doCheckSpeedLimit(superFunc)
    if superFunc ~= nil then
        return superFunc(self) or (self:getIsTurnedOn() and self:allowPickingUp());
    end;
    return self:getIsTurnedOn() and self:allowPickingUp();
end;

function Ultima:setUnitFillLevel(superFunc, fillUnitIndex, fillLevel, fillType, force, fillInfo)
    superFunc(self, fillUnitIndex, fillLevel, fillType, force, fillInfo);    
    if fillUnitIndex == self.preChamberUnitIndex then
        self.prePressChamberText:setText(string.gsub(g_i18n:getText("prePressChamber"), "percent" , floor2(fillLevel/self:getUnitCapacity(fillUnitIndex))));
    elseif fillUnitIndex == self.mainChamberUnitIndex then        
        self.mainPressChamberText:setText(string.gsub(g_i18n:getText("mainPressChamber"), "percent" , floor2(fillLevel/self:getUnitCapacity(fillUnitIndex))));
    
        if self.currentBalerState == Ultima.STATE_usualLoading or self.currentBalerState == Ultima.STATE_passLoad then
            if self.balerCurrentBale == nil and fillLevel > 0 and fillType ~= FillUtil.FILLTYPE_UNKNOWN then
                -- balerCreateBale is not synched at all
                self.balerCurrentBale = self:balerCreateBale(self.baleSizeIndex, fillType);
                self:setUnitCapacity(self.mainChamberUnitIndex, self.balerCurrentBale.fillLevel);
                self.chamberFillAnim:play();
            end;
            local animTime = (fillLevel / self:getUnitCapacity(fillUnitIndex)) * (self.baleSizes[self.baleSizeIndex].chamberAnimTime / self.chamberFillAnim.duration);
            self.chamberFillAnim:setRelativeTime(animTime);
        end;
    end;
end;

function Ultima:allowFillType(fillType)
    return self:getUnitFillTypes(self.mainChamberUnitIndex)[fillType] ~= nil;
end;

function Ultima:setPickupState(oldFunc, isPickupLowered, noEventSend)
    oldFunc(self, isPickupLowered, noEventSend);
    if self:getIsActiveForSound() then
        if isPickupLowered then
            SoundUtil.playSample(self.pickupDownSound, 1, 0, nil);
        else
            SoundUtil.playSample(self.pickupUpSound, 1, 0, nil);
        end;
    end;
end;

function Ultima:setBaleCounter(day, lifeTime, noEventSend)     -- note: usually triggered by server, only reset triggered by client
    self.baleCountDay = day;
    self.baleCountLifetime = lifeTime;
    self.baleCounterSingleText:setText(tostring(day));
    self.baleCounterAllText:setText(tostring(lifeTime));
    
    if not noEventSend then
        SetBalerCountEvent.sendEvent(self, day, lifeTime);
    end;
end;

function Ultima:getDefaultSpeedLimit() -- to enable speed limit check
    return self.maxSpeedLimit;
end;

function Ultima:getDirtMultiplier(oldFunc)
    local multiplier = 0;
    if oldFunc ~= nil then
        multiplier = oldFunc(self);
    end;
    if self:getIsTurnedOn() then
        multiplier = self.workMultiplier;
    end;
    return multiplier;
end;

function Ultima.getRefillRolesSingleItemLoad(xmlFile, baseXMLName, baseDir, customEnvironment, isMod, configItem)
    configItem.manualRefill = Utils.getNoNil(getXMLBool(xmlFile, baseXMLName.."#manualRefill"), false);
end;

ConfigurationUtil.registerConfigurationType('refillNetRoles', g_i18n:getText('configuration_refillNetRoles'), nil, Ultima.getRefillRolesSingleItemLoad, nil, ConfigurationUtil.SELECTOR_MULTIOPTION); -- Vehicle.getConfigColorPostLoad
ConfigurationUtil.registerConfigurationType('refillFoilRoles', g_i18n:getText('configuration_refillFoilRoles'), nil, Ultima.getRefillRolesSingleItemLoad, nil, ConfigurationUtil.SELECTOR_MULTIOPTION);